#!/usr/bin/python

# Copyright (c) 2024, Stanislav Shamilov <shamilovstas@protonmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: android_sdk
short_description: Manages Android SDK packages
description:
  - Manages Android SDK packages.
  - Allows installation from different channels (stable, beta, dev, canary).
  - Allows installation of packages to a non-default SDK root directory.
author: Stanislav Shamilov (@shamilovstas)
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
version_added: 10.2.0
options:
  accept_licenses:
    description:
      - If this is set to V(true), the module attempts to accept license prompts generated by C(sdkmanager) during package
        installation. Otherwise, every license prompt is rejected.
    type: bool
    default: false
  name:
    description:
      - A name of an Android SDK package (for instance, V(build-tools;34.0.0)).
    aliases: ['package', 'pkg']
    type: list
    elements: str
  state:
    description:
      - Indicates the desired package(s) state.
      - V(present) ensures that package(s) is/are present.
      - V(absent) ensures that package(s) is/are absent.
      - V(latest) ensures that package(s) is/are installed and updated to the latest version(s).
    choices: ['present', 'absent', 'latest']
    default: present
    type: str
  sdk_root:
    description:
      - Provides path for an alternative directory to install Android SDK packages to. By default, all packages are installed
        to the directory where C(sdkmanager) is installed.
    type: path
  channel:
    description:
      - Indicates what channel must C(sdkmanager) use for installation of packages.
    choices: ['stable', 'beta', 'dev', 'canary']
    default: stable
    type: str
requirements:
  - C(java) >= 17
  - C(sdkmanager) Command line tool for installing Android SDK packages.
notes:
  - For some of the packages installed by C(sdkmanager) is it necessary to accept licenses. Usually it is done through command
    line prompt in a form of a Y/N question when a licensed package is requested to be installed. If there are several packages
    requested for installation and at least two of them belong to different licenses, the C(sdkmanager) tool prompts for these
    licenses in a loop. In order to install packages, the module must be able to answer these license prompts. Currently,
    it is only possible to answer one license prompt at a time, meaning that instead of installing multiple packages as a
    single invocation of the C(sdkmanager --install) command, it is done by executing the command independently for each package.
    This makes sure that at most only one license prompt needs to be answered. At the time of writing this module, a C(sdkmanager)'s
    package may belong to at most one license type that needs to be accepted. However, if this changes in the future, the
    module may hang as there might be more prompts generated by the C(sdkmanager) tool which the module is unable to answer.
    If this becomes the case, file an issue and in the meantime, consider accepting all the licenses in advance, as it is
    described in the C(sdkmanager) L(documentation,https://developer.android.com/tools/sdkmanager#accept-licenses), for instance,
    using the M(ansible.builtin.command) module.
seealso:
  - name: sdkmanager tool documentation
    description: Detailed information of how to install and use sdkmanager command line tool.
    link: https://developer.android.com/tools/sdkmanager
"""

EXAMPLES = r"""
- name: Install build-tools;34.0.0
  community.general.android_sdk:
    name: build-tools;34.0.0
    accept_licenses: true
    state: present

- name: Install build-tools;34.0.0 and platform-tools
  community.general.android_sdk:
    name:
      - build-tools;34.0.0
      - platform-tools
    accept_licenses: true
    state: present

- name: Delete build-tools;34.0.0
  community.general.android_sdk:
    name: build-tools;34.0.0
    state: absent

- name: Install platform-tools or update if installed
  community.general.android_sdk:
    name: platform-tools
    accept_licenses: true
    state: latest

- name: Install build-tools;34.0.0 to a different SDK root
  community.general.android_sdk:
    name: build-tools;34.0.0
    accept_licenses: true
    state: present
    sdk_root: "/path/to/new/root"

- name: Install a package from another channel
  community.general.android_sdk:
    name: some-package-present-in-canary-channel
    accept_licenses: true
    state: present
    channel: canary
"""

RETURN = r"""
installed:
  description: A list of packages that have been installed.
  returned: when packages have changed
  type: list
  sample: ["build-tools;34.0.0", "platform-tools"]

removed:
  description: A list of packages that have been removed.
  returned: when packages have changed
  type: list
  sample: ["build-tools;34.0.0", "platform-tools"]
"""

from ansible_collections.community.general.plugins.module_utils._android_sdkmanager import AndroidSdkManager, Package
from ansible_collections.community.general.plugins.module_utils._mh.module_helper import StateModuleHelper


class AndroidSdk(StateModuleHelper):
    module = dict(
        argument_spec=dict(
            state=dict(type="str", default="present", choices=["present", "absent", "latest"]),
            package=dict(type="list", elements="str", aliases=["pkg", "name"]),
            sdk_root=dict(type="path"),
            channel=dict(type="str", default="stable", choices=["stable", "beta", "dev", "canary"]),
            accept_licenses=dict(type="bool", default=False),
        ),
        supports_check_mode=True,
    )

    def __init_module__(self):
        self.sdkmanager = AndroidSdkManager(self.module)
        self.vars.set("installed", [], change=True)
        self.vars.set("removed", [], change=True)

    def _parse_packages(self):
        arg_pkgs = set(self.vars.package)
        if len(arg_pkgs) < len(self.vars.package):
            self.do_raise("Packages may not repeat")
        return {Package(p) for p in arg_pkgs}

    def state_present(self):
        packages = self._parse_packages()
        installed = self.sdkmanager.get_installed_packages()
        pending_installation = packages.difference(installed)

        self.vars.installed = AndroidSdk._map_packages_to_names(pending_installation)
        if not self.check_mode:
            rc, stdout, stderr = self.sdkmanager.apply_packages_changes(pending_installation, self.vars.accept_licenses)
            if rc != 0:
                self.do_raise(f"Could not install packages: {stderr}")

    def state_absent(self):
        packages = self._parse_packages()
        installed = self.sdkmanager.get_installed_packages()
        to_be_deleted = packages.intersection(installed)
        self.vars.removed = AndroidSdk._map_packages_to_names(to_be_deleted)
        if not self.check_mode:
            rc, stdout, stderr = self.sdkmanager.apply_packages_changes(to_be_deleted)
            if rc != 0:
                self.do_raise(f"Could not uninstall packages: {stderr}")

    def state_latest(self):
        packages = self._parse_packages()
        installed = self.sdkmanager.get_installed_packages()
        updatable = self.sdkmanager.get_updatable_packages()
        not_installed = packages.difference(installed)
        to_be_installed = not_installed.union(updatable)
        self.vars.installed = AndroidSdk._map_packages_to_names(to_be_installed)

        if not self.check_mode:
            rc, stdout, stderr = self.sdkmanager.apply_packages_changes(to_be_installed, self.vars.accept_licenses)
            if rc != 0:
                self.do_raise(f"Could not install packages: {stderr}")

    @staticmethod
    def _map_packages_to_names(packages):
        return [x.name for x in packages]


def main():
    AndroidSdk.execute()


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2025 Klaas Demter (@Klaas-)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_monitorscheduledqueryrules_info
version_added: "3.11.0"
short_description: Get Scheduled query rules
description:
    - Get Scheduled query rules
    - These rules are also called Log search alert rules

options:
    name:
        description:
            - The name of the scheduled query rule you're trying to get details about.
        type: str
    resource_group:
        description:
            - The name of the resource group in which the scheduled query rule is (if you use name)
            - The name of the resource group where you want to list Scheduled query rules (if you don't use name)
        type: str
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Klaas Demter (@Klaas-)
'''

EXAMPLES = '''
- name: Get scheduled query rule details
  azure.azcollection.azure_rm_monitorscheduledqueryrules_info:
    name: scheduled_query_rule_name
    resource_group: resource_group_name

- name: Get all Scheduled query rules in specific resource group
  azure.azcollection.azure_rm_monitorscheduledqueryrules_info:
    resource_group: resource_group_name

- name: Get all Scheduled query rules in the current subscription
  azure.azcollection.azure_rm_monitorscheduledqueryrules_info:
'''

RETURN = '''
scheduledqueryrules:
    description:
        - List of Scheduled query rules
        - Can be empty if listing Scheduled query rules
    type: list
    returned: always
    sample: [
        {
            "action": {
                "azns_action": {
                    "action_group": [
                        "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name/providers/microsoft.insights/actionGroups/action_group_name"
                    ],
                    "email_subject": ""
                },
                "odata_type": \
"Microsoft.WindowsAzure.Management.Monitoring.Alerts.Models.Microsoft.AppInsights.Nexus.DataContracts.Resources.ScheduledQueryRules.AlertingAction",
                "severity": "2",
                "trigger": {
                    "metric_trigger": {
                        "metric_column": "InstanceName,_ResourceId",
                        "metric_trigger_type": "Total",
                        "threshold": 1.0,
                        "threshold_operator": "GreaterThanOrEqual"
                    },
                    "threshold": 10.0,
                    "threshold_operator": "LessThan"
                }
            },
            "auto_mitigate": true,
            "created_with_api_version": "2018-04-16",
            "description": "Monitoring: Disks/logical volumes with less than 10% free space",
            "display_name": "scheduled_query_alert_name",
            "enabled": "true",
            "id": \
"/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name/providers/microsoft.insights/scheduledqueryrules/scheduled_query_alert_name",
            "last_updated_time": "2025-10-30T16:07:30.773011Z",
            "location": "eastus",
            "name": "scheduled_query_alert_name",
            "provisioning_state": "Succeeded",
            "schedule": {
                "frequency_in_minutes": 5,
                "time_window_in_minutes": 5
            },
            "source": {
                "authorized_resources": [],
                "data_source_id": \
"/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name/providers/Microsoft.OperationalInsights/workspaces/log_analytics_workspace_names",
                "query": "actual query",
                "query_type": "ResultCount"
            },
            "tags": {},
            "type": "Microsoft.Insights/scheduledQueryRules"
        }
    ]
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.core.exceptions import ResourceNotFoundError, HttpResponseError

except ImportError:
    # This is handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'scheduledqueryrules'


class AzureRMscheduledqueryrulesInfo(AzureRMModuleBase):
    """Information class for an Azure RM Scheduled query rules"""

    def __init__(self):
        self.module_arg_spec = dict(
            name=dict(type='str', required=False),
            resource_group=dict(type='str', required=False),
        )

        self.required_by = {
            'name': 'resource_group'
        }

        self.name = None
        self.resource_group = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            changed=False,
            scheduledqueryrules=[]
        )

        super(AzureRMscheduledqueryrulesInfo, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                             supports_check_mode=True,
                                                             supports_tags=False,
                                                             facts_module=True,
                                                             required_by=self.required_by)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name:
            result = self.get_scheduled_query_rule()
        else:
            result = self.list_scheduled_query_rules()

        self.results['scheduledqueryrules'] = result

        return self.results

    def get_scheduled_query_rule(self):
        '''
        Gets the properties of the specified scheduled query rule.

        :return: List of Scheduled query rules
        '''
        self.log("Checking if scheduled query rule {0} in resource group {1} is present".format(self.name,
                                                                                                self.resource_group))

        result = []
        scheduled_query_rule = None

        try:
            scheduled_query_rule = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules.get(rule_name=self.name,
                                                                                                                  resource_group_name=self.resource_group)
        except ResourceNotFoundError as ex:
            self.log("Could not find scheduled query rule {0} in resource group {1}".format(self.name, self.resource_group))
            return []
        except HttpResponseError as ex:
            if ex.error.code == 'InvalidSubscriptionId':
                self.log("Could not find subscription id")
                return []
            else:
                raise Exception(ex)
        if scheduled_query_rule:
            result = [self.serialize_obj(scheduled_query_rule, AZURE_OBJECT_CLASS)]

        return result

    def list_scheduled_query_rules(self):
        '''
        Gets the properties of the specified Scheduled query rules in resource group or subscription.

        :return: List of Scheduled query rules
        '''

        result = []
        scheduled_query_rules = None

        if self.resource_group:
            self.log("Checking if the Scheduled query rules in resource group {0} are present".format(self.resource_group))
            scheduled_query_rules_mgmt_client = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules
            scheduled_query_rules = scheduled_query_rules_mgmt_client.list_by_resource_group(resource_group_name=self.resource_group)
        else:
            self.log("Checking if the Scheduled query rules are present in subscription")
            scheduled_query_rules = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules.list_by_subscription()
        if scheduled_query_rules:
            # it seems the exception is thrown when iterating through scheduled_query_rules, not when setting scheduled_query_rules
            try:
                for item in scheduled_query_rules:
                    result.append(self.serialize_obj(item, AZURE_OBJECT_CLASS))
            except ResourceNotFoundError as ex:
                self.log("Could not find resource group {0}".format(self.resource_group))
            except HttpResponseError as ex:
                if ex.error.code == 'InvalidSubscriptionId':
                    self.log("Could not find subscription id")
                else:
                    raise Exception(ex)

        return result


def main():
    """Main execution"""
    AzureRMscheduledqueryrulesInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Ansible module to manage CheckPoint Firewall (c) 2019
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import (absolute_import, division, print_function)

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = """
---
module: cp_mgmt_show_software_packages_per_targets
short_description: Shows software packages on targets.
description:
  - Shows software packages on targets.
  - All operations are performed over Web Services API.
  - Available from R80.40 management version.
version_added: "5.0.0"
author: "Shiran Golzar (@chkp-shirango)"
options:
  display:
    description:
      - Filter the displayed results.
    type: dict
    suboptions:
      category:
        description:
          - The package categories to include in the results.
        type: list
        elements: str
      installed:
        description:
          - Show installed packages, available packages, or both.
        type: str
        choices: ['yes', 'no', 'any']
      recommended:
        description:
          - Show only recommended packages, other packages, or both.
        type: str
        choices: ['yes', 'no', 'any']
  targets:
    description:
      - On what targets to execute this command. Targets may be identified by their object name, or object unique identifier.
    type: list
    elements: str
extends_documentation_fragment: check_point.mgmt.checkpoint_commands
"""

EXAMPLES = """
- name: show-software-packages-per-targets
  cp_mgmt_show_software_packages_per_targets:
    display:
      category: major
      installed: 'no'
      recommended: any
    targets:
      - corporate-gateway
"""

RETURN = """
cp_mgmt_show_software_packages_per_targets:
  description: The checkpoint show-software-packages-per-targets output.
  returned: always.
  type: dict
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import checkpoint_argument_spec_for_commands, \
    api_command


def main():
    argument_spec = dict(
        display=dict(type='dict', options=dict(
            category=dict(type='list', elements='str'),
            installed=dict(type='str', choices=['yes', 'no', 'any']),
            recommended=dict(type='str', choices=['yes', 'no', 'any'])
        )),
        targets=dict(type='list', elements='str')
    )
    argument_spec.update(checkpoint_argument_spec_for_commands)

    module = AnsibleModule(argument_spec=argument_spec)

    command = "show-software-packages-per-targets"

    result = api_command(module, command)
    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_camera_wireless_profiles
short_description: Resource module for networks _camera _wireless _profiles
description:
  - Manage operations create, update and delete of the resource networks _camera _wireless
    _profiles.
  - Creates a new camera wireless profile for this network.
  - Delete an existing camera wireless profile for this network.
  - Update an existing camera wireless profile in this network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  identity:
    description: The identity of the wireless profile. Required for creating wireless
      profiles in 8021x-radius auth mode.
    suboptions:
      password:
        description: The password of the identity.
        type: str
      username:
        description: The username of the identity.
        type: str
    type: dict
  name:
    description: The name of the camera wireless profile. This parameter is required.
    type: str
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  ssid:
    description: The details of the SSID config.
    suboptions:
      authMode:
        description: The auth mode of the SSID. It can be set to ('psk', '8021x-radius').
        type: str
      encryptionMode:
        description: The encryption mode of the SSID. It can be set to ('wpa', 'wpa-eap').
          With 'wpa' mode, the authMode should be 'psk' and with 'wpa-eap' the authMode
          should be '8021x-radius'.
        type: str
      name:
        description: The name of the SSID.
        type: str
      psk:
        description: The pre-shared key of the SSID.
        type: str
    type: dict
  wirelessProfileId:
    description: WirelessProfileId path parameter. Wireless profile ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for camera createNetworkCameraWirelessProfile
    description: Complete reference of the createNetworkCameraWirelessProfile API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-network-camera-wireless-profile
  - name: Cisco Meraki documentation for camera deleteNetworkCameraWirelessProfile
    description: Complete reference of the deleteNetworkCameraWirelessProfile API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-network-camera-wireless-profile
  - name: Cisco Meraki documentation for camera updateNetworkCameraWirelessProfile
    description: Complete reference of the updateNetworkCameraWirelessProfile API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-camera-wireless-profile
notes:
  - SDK Method used are
    camera.Camera.create_network_camera_wireless_profile,
    camera.Camera.delete_network_camera_wireless_profile,
    camera.Camera.update_network_camera_wireless_profile,
  - Paths used are
    post /networks/{networkId}/camera/wirelessProfiles,
    delete /networks/{networkId}/camera/wirelessProfiles/{wirelessProfileId},
    put /networks/{networkId}/camera/wirelessProfiles/{wirelessProfileId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_camera_wireless_profiles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    identity:
      password: password123
      username: identityname
    name: wireless profile A
    networkId: string
    ssid:
      authMode: 8021x-radius
      encryptionMode: wpa-eap
      name: ssid test
      psk: sampleKey
- name: Update by id
  cisco.meraki.networks_camera_wireless_profiles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    identity:
      password: password123
      username: identityname
    name: wireless profile A
    networkId: string
    ssid:
      authMode: 8021x-radius
      encryptionMode: wpa-eap
      name: ssid test
      psk: sampleKey
    wirelessProfileId: string
- name: Delete by id
  cisco.meraki.networks_camera_wireless_profiles:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    networkId: string
    wirelessProfileId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "name": "string",
      "appliedDeviceCount": 0,
      "ssid": {
        "name": "string",
        "authMode": "string",
        "encryptionMode": "string",
        "psk": "string"
      },
      "identity": {
        "username": "string",
        "password": "string"
      }
    }
"""



# Copyright (c) 2022, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import os
import typing as t

from ansible_collections.community.general.plugins.module_utils._cmd_runner import CmdRunner, cmd_runner_fmt

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


_PUPPET_PATH_PREFIX = ["/opt/puppetlabs/bin"]


def get_facter_dir() -> str:
    if os.getuid() == 0:
        return "/etc/facter/facts.d"
    else:
        return os.path.expanduser("~/.facter/facts.d")


def _puppet_cmd(module: AnsibleModule) -> str | None:
    return module.get_bin_path("puppet", False, _PUPPET_PATH_PREFIX)


# If the `timeout` CLI command feature is removed,
# Then we could add this as a fixed param to `puppet_runner`
def ensure_agent_enabled(module: AnsibleModule) -> None:
    runner = CmdRunner(
        module,
        command="puppet",
        path_prefix=_PUPPET_PATH_PREFIX,
        arg_formats=dict(
            _agent_disabled=cmd_runner_fmt.as_fixed(["config", "print", "agent_disabled_lockfile"]),
        ),
        check_rc=False,
    )

    rc, stdout, stderr = runner("_agent_disabled").run()
    if os.path.exists(stdout.strip()):
        module.fail_json(msg="Puppet agent is administratively disabled.", disabled=True)
    elif rc != 0:
        module.fail_json(msg="Puppet agent state could not be determined.")


def puppet_runner(module: AnsibleModule) -> CmdRunner:
    # Keeping backward compatibility, allow for running with the `timeout` CLI command.
    # If this can be replaced with ansible `timeout` parameter in playbook,
    # then this function could be removed.
    def _prepare_base_cmd():
        if module.params["timeout"]:
            _tout_cmd = module.get_bin_path("timeout", False)
            if _tout_cmd:
                return ["timeout", "-s", "9", module.params["timeout"], _puppet_cmd(module)]
        return ["puppet"]

    def noop_func(v):
        return ["--noop"] if module.check_mode or v else ["--no-noop"]

    _logdest_map = {
        "syslog": ["--logdest", "syslog"],
        "all": ["--logdest", "syslog", "--logdest", "console"],
    }

    @cmd_runner_fmt.unpack_args
    def execute_func(execute, manifest):
        if execute:
            return ["--execute", execute]
        else:
            return [manifest]

    runner = CmdRunner(
        module,
        command=_prepare_base_cmd(),
        path_prefix=_PUPPET_PATH_PREFIX,
        arg_formats=dict(
            _agent_fixed=cmd_runner_fmt.as_fixed(
                [
                    "agent",
                    "--onetime",
                    "--no-daemonize",
                    "--no-usecacheonfailure",
                    "--no-splay",
                    "--detailed-exitcodes",
                    "--verbose",
                    "--color",
                    "0",
                ]
            ),
            _apply_fixed=cmd_runner_fmt.as_fixed(["apply", "--detailed-exitcodes"]),
            puppetmaster=cmd_runner_fmt.as_opt_val("--server"),
            show_diff=cmd_runner_fmt.as_bool("--show-diff"),
            confdir=cmd_runner_fmt.as_opt_val("--confdir"),
            environment=cmd_runner_fmt.as_opt_val("--environment"),
            tags=cmd_runner_fmt.as_func(lambda v: ["--tags", ",".join(v)]),
            skip_tags=cmd_runner_fmt.as_func(lambda v: ["--skip_tags", ",".join(v)]),
            certname=cmd_runner_fmt.as_opt_eq_val("--certname"),
            noop=cmd_runner_fmt.as_func(noop_func),
            use_srv_records=cmd_runner_fmt.as_bool("--usr_srv_records", "--no-usr_srv_records", ignore_none=True),
            logdest=cmd_runner_fmt.as_map(_logdest_map, default=[]),
            modulepath=cmd_runner_fmt.as_opt_eq_val("--modulepath"),
            _execute=cmd_runner_fmt.as_func(execute_func),
            summarize=cmd_runner_fmt.as_bool("--summarize"),
            waitforlock=cmd_runner_fmt.as_opt_val("--waitforlock"),
            debug=cmd_runner_fmt.as_bool("--debug"),
            verbose=cmd_runner_fmt.as_bool("--verbose"),
        ),
        check_rc=False,
        force_lang=module.params["environment_lang"],
    )
    return runner



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_vpc_net
version_added: 1.0.0
short_description: Configure AWS Virtual Private Clouds
description:
  - Create, modify, and terminate AWS Virtual Private Clouds (VPCs).
author:
  - Jonathan Davila (@defionscode)
  - Sloane Hertel (@s-hertel)
options:
  name:
    description:
      - The name to give your VPC. This is used in combination with O(cidr_block)
        to determine if a VPC already exists.
      - The value of O(name) overrides any value set for V(Name) in the O(tags)
        parameter.
      - At least one of O(name) and O(vpc_id) must be specified.
      - O(name) must be specified when creating a new VPC.
    type: str
  vpc_id:
    version_added: 4.0.0
    description:
      - The ID of the VPC.
      - At least one of O(name) and O(vpc_id) must be specified.
      - At least one of O(name) and O(cidr_block) must be specified.
    type: str
  cidr_block:
    description:
      - The primary CIDR of the VPC.
      - The first in the list will be used as the primary CIDR
        and is used in conjunction with O(name) to ensure idempotence.
      - Required when O(vpc_id) is not set.
    type: list
    elements: str
  ipv6_cidr:
    description:
      - Request an Amazon-provided IPv6 CIDR block with /56 prefix length. You cannot specify the range of IPv6 addresses,
        or the size of the CIDR block.
      - Default value is V(false) when creating a new VPC.
    type: bool
  purge_cidrs:
    description:
      - Remove CIDRs that are associated with the VPC and are not specified in O(cidr_block).
    default: false
    type: bool
  tenancy:
    description:
      - Whether to be V(default) or V(dedicated) tenancy.
      - This cannot be changed after the VPC has been created.
    default: default
    choices: [ 'default', 'dedicated' ]
    type: str
  dns_support:
    description:
      - Whether to enable AWS DNS support.
      - Default value is V(true) when creating a new VPC.
    type: bool
  dns_hostnames:
    description:
      - Whether to enable AWS hostname support.
      - Default value is V(true) when creating a new VPC.
    type: bool
  dhcp_opts_id:
    description:
      - The id of the DHCP options to use for this VPC.
    type: str
  state:
    description:
      - The state of the VPC. Either absent or present.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  multi_ok:
    description:
      - By default the module will not create another VPC if there is another VPC with the same name and CIDR block.
        Specify O(multi_ok=true) if you want duplicate VPCs created.
    type: bool
    default: false
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: create a VPC with dedicated tenancy and a couple of tags
  amazon.aws.ec2_vpc_net:
    name: Module_dev2
    cidr_block: 10.10.0.0/16
    region: us-east-1
    tags:
      module: ec2_vpc_net
      this: works
    tenancy: dedicated

- name: create a VPC with dedicated tenancy and request an IPv6 CIDR
  amazon.aws.ec2_vpc_net:
    name: Module_dev2
    cidr_block: 10.10.0.0/16
    ipv6_cidr: true
    region: us-east-1
    tenancy: dedicated

- name: Delete an existing VPC
  amazon.aws.ec2_vpc_net:
    vpc_id: vpc-0123456789abcdef0
    state: absent
"""

RETURN = r"""
vpc:
  description: Info about the VPC that was created or deleted.
  returned: always
  type: complex
  contains:
    cidr_block:
      description: The CIDR of the VPC.
      returned: always
      type: str
      sample: 10.0.0.0/16
    cidr_block_association_set:
      description: IPv4 CIDR blocks associated with the VPC.
      returned: success
      type: list
      sample:
        "cidr_block_association_set": [
            {
                "association_id": "vpc-cidr-assoc-97aeeefd",
                "cidr_block": "10.0.0.0/24",
                "cidr_block_state": {
                    "state": "associated"
                }
            }
        ]
    dhcp_options_id:
      description: The id of the DHCP options associated with this VPC.
      returned: always
      type: str
      sample: dopt-12345678
    id:
      description: VPC resource id.
      returned: always
      type: str
      sample: vpc-12345678
    name:
      description: The Name tag of the VPC.
      returned: When the Name tag has been set on the VPC
      type: str
      sample: MyVPC
      version_added: 4.0.0
    instance_tenancy:
      description: Indicates whether VPC uses default or dedicated tenancy.
      returned: always
      type: str
      sample: default
    ipv6_cidr_block_association_set:
      description: IPv6 CIDR blocks associated with the VPC.
      returned: success
      type: list
      sample:
        "ipv6_cidr_block_association_set": [
            {
                "association_id": "vpc-cidr-assoc-97aeeefd",
                "ipv6_cidr_block": "2001:db8::/56",
                "ipv6_cidr_block_state": {
                    "state": "associated"
                }
            }
        ]
    is_default:
      description: Indicates whether this is the default VPC.
      returned: always
      type: bool
      sample: false
    state:
      description: State of the VPC.
      returned: always
      type: str
      sample: available
    tags:
      description: Tags attached to the VPC, includes name.
      returned: always
      type: complex
      contains:
        Name:
          description: Name tag for the VPC.
          returned: always
          type: str
          sample: pk_vpc4
    owner_id:
      description: The AWS account which owns the VPC.
      returned: always
      type: str
      sample: 123456789012
"""

from time import sleep
from time import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict
from ansible.module_utils.common.network import to_subnet

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import associate_dhcp_options
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import associate_vpc_cidr_block
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_vpc
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_vpc
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_vpc_attribute
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_vpcs
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import disassociate_vpc_cidr_block
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import modify_vpc_attribute
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list
from ansible_collections.amazon.aws.plugins.module_utils.waiters import wait_for_resource_state


def get_vpc_id(connection, module: AnsibleAWSModule) -> Optional[str]:
    """Returns None or a vpc object depending on the existence of a VPC. When supplied
    with a CIDR, it will check for matching tags to determine if it is a match
    otherwise it will assume the VPC does not exist and thus return None.
    """
    name = module.params.get("name")
    cidr_block = module.params.get("cidr_block")
    multi_ok = module.params.get("multi_ok")
    vpc_filters = ansible_dict_to_boto3_filter_list({"tag:Name": name, "cidr-block": cidr_block})
    matching_vpcs = describe_vpcs(connection, Filters=vpc_filters)
    # If an exact matching using a list of CIDRs isn't found, check for a match with the first CIDR as is documented for C(cidr_block)
    if not matching_vpcs:
        vpc_filters = ansible_dict_to_boto3_filter_list({"tag:Name": name, "cidr-block": [cidr_block[0]]})
        matching_vpcs = describe_vpcs(connection, Filters=vpc_filters)

    vpc_id = None
    if not multi_ok and matching_vpcs:
        if len(matching_vpcs) > 1:
            module.fail_json(
                msg=(
                    f"Currently there are {len(matching_vpcs)} VPCs that have the same name and CIDR block you specified."
                    " If you would like to create the VPC anyway please pass True to the multi_ok param."
                )
            )
        vpc_id = matching_vpcs[0]["VpcId"]
    return vpc_id


def get_vpc(module: AnsibleAWSModule, connection, vpc_id: str) -> Dict[str, Any]:
    wait_for_resource_state(connection, module, "vpc_available", VpcIds=[vpc_id])
    return describe_vpcs(connection, VpcIds=[vpc_id])[0]


def update_vpc_tags(connection, module, vpc_id, tags, name, purge_tags):
    # Name is a tag rather than a direct parameter, we need to inject 'Name'
    # into tags, but since tags isn't explicitly passed we'll treat it not being
    # set as purge_tags == False
    if name:
        if purge_tags and tags is None:
            purge_tags = False
        tags = tags or {}
        tags.update({"Name": name})

    if tags is None:
        return False

    changed = ensure_ec2_tags(connection, module, vpc_id, tags=tags, purge_tags=purge_tags)
    if not changed or module.check_mode:
        return changed

    return True


def update_dhcp_opts(connection, module: AnsibleAWSModule, vpc_obj: Dict[str, Any], dhcp_id: Optional[str]) -> bool:
    if dhcp_id is None:
        return False
    if vpc_obj["DhcpOptionsId"] == dhcp_id:
        return False
    if module.check_mode:
        return True

    return associate_dhcp_options(connection, vpc_id=vpc_obj["VpcId"], dhcp_options_id=dhcp_id)


def create_vpc_net(
    connection,
    module: AnsibleAWSModule,
    cidr_block: List[str],
    tenancy,
    tags: Dict[str, str],
    ipv6_cidr: bool,
    name: Optional[str],
) -> Dict[str, Any]:
    if module.check_mode:
        module.exit_json(changed=True, msg="VPC would be created if not in check mode")

    create_args = dict(
        CidrBlock=cidr_block,
        InstanceTenancy=tenancy,
    )

    if name:
        tags = tags or {}
        tags["Name"] = name
    if tags:
        create_args["TagSpecifications"] = boto3_tag_specifications(tags, "vpc")

    # Defaults to False (including None)
    if ipv6_cidr:
        create_args["AmazonProvidedIpv6CidrBlock"] = True

    vpc_obj = create_vpc(connection, **create_args)

    # Wait up to 30 seconds for vpc to exist and for state 'available'
    for waiter_name in ("vpc_exists", "vpc_available"):
        wait_for_resource_state(connection, module, waiter_name, max_attempts=30, VpcIds=[vpc_obj["VpcId"]])

    return vpc_obj


def wait_for_vpc_attribute(connection, module, vpc_id, attribute, expected_value):
    if expected_value is None:
        return
    if module.check_mode:
        return

    start_time = time()
    updated = False
    while time() < start_time + 300:
        current_value = describe_vpc_attribute(connection, attribute=attribute, vpc_id=vpc_id)[
            f"{attribute[0].upper()}{attribute[1:]}"
        ]["Value"]
        if current_value != expected_value:
            sleep(3)
        else:
            updated = True
            break
    if not updated:
        module.fail_json(msg=f"Failed to wait for {attribute} to be updated")


def wait_for_vpc_ipv6_state(module, connection, vpc_id, ipv6_assoc_state):
    """
    If ipv6_assoc_state is True, wait for VPC to be associated with at least one Amazon-provided IPv6 CIDR block.
    If ipv6_assoc_state is False, wait for VPC to be dissociated from all Amazon-provided IPv6 CIDR blocks.
    """

    if ipv6_assoc_state is None:
        return
    if module.check_mode:
        return

    start_time = time()
    criteria_match = False
    while time() < start_time + 300:
        current_value = get_vpc(module, connection, vpc_id)
        if current_value:
            ipv6_set = current_value.get("Ipv6CidrBlockAssociationSet")
            # "ipv6_cidr": false and no Ipv6CidrBlockAssociationSet
            if not ipv6_set and not ipv6_assoc_state:
                return
            if ipv6_set:
                if ipv6_assoc_state:
                    # At least one 'Amazon' IPv6 CIDR block must be associated.
                    for val in ipv6_set:
                        if (
                            val.get("Ipv6Pool") == "Amazon"
                            and val.get("Ipv6CidrBlockState").get("State") == "associated"
                        ):
                            criteria_match = True
                            break
                    if criteria_match:
                        break
                else:
                    # All 'Amazon' IPv6 CIDR blocks must be disassociated.
                    expected_count = sum([(val.get("Ipv6Pool") == "Amazon") for val in ipv6_set])
                    actual_count = sum(
                        [
                            (
                                val.get("Ipv6Pool") == "Amazon"
                                and val.get("Ipv6CidrBlockState").get("State") == "disassociated"
                            )
                            for val in ipv6_set
                        ]
                    )
                    if actual_count == expected_count:
                        criteria_match = True
                        break
        sleep(3)
    if not criteria_match:
        module.fail_json(msg="Failed to wait for IPv6 CIDR association")


def get_cidr_network_bits(module, cidr_block):
    if cidr_block is None:
        return None

    fixed_cidrs = []
    for cidr in cidr_block:
        split_addr = cidr.split("/")
        if len(split_addr) == 2:
            # this_ip is a IPv4 CIDR that may or may not have host bits set
            # Get the network bits.
            valid_cidr = to_subnet(split_addr[0], split_addr[1])
            if cidr != valid_cidr:
                module.warn(
                    f"One of your CIDR addresses ({cidr}) has host bits set. To get rid of this warning, check the"
                    f" network mask and make sure that only network bits are set: {valid_cidr}."
                )
            fixed_cidrs.append(valid_cidr)
        else:
            # let AWS handle invalid CIDRs
            fixed_cidrs.append(cidr)
    return fixed_cidrs


def update_ipv6_cidrs(connection, module, vpc_obj, vpc_id, ipv6_cidr):
    if ipv6_cidr is None:
        return False

    # Fetch current state from vpc_object
    current_ipv6_cidr = False
    if "Ipv6CidrBlockAssociationSet" in vpc_obj.keys():
        for ipv6_assoc in vpc_obj["Ipv6CidrBlockAssociationSet"]:
            if ipv6_assoc["Ipv6Pool"] == "Amazon" and ipv6_assoc["Ipv6CidrBlockState"]["State"] in [
                "associated",
                "associating",
            ]:
                current_ipv6_cidr = True
                break

    if ipv6_cidr == current_ipv6_cidr:
        return False

    if module.check_mode:
        return True

    # There's no block associated, and we want one to be associated
    if ipv6_cidr:
        associate_vpc_cidr_block(connection, vpc_id=vpc_id, AmazonProvidedIpv6CidrBlock=ipv6_cidr)
    else:
        for ipv6_assoc in vpc_obj["Ipv6CidrBlockAssociationSet"]:
            if ipv6_assoc["Ipv6Pool"] == "Amazon" and ipv6_assoc["Ipv6CidrBlockState"]["State"] in [
                "associated",
                "associating",
            ]:
                disassociate_vpc_cidr_block(connection, association_id=ipv6_assoc["AssociationId"])
    return True


def update_cidrs(connection, module, vpc_obj, cidr_block, purge_cidrs):
    if not cidr_block:
        return False, None

    associated_cidrs = dict(
        (cidr["CidrBlock"], cidr["AssociationId"])
        for cidr in vpc_obj.get("CidrBlockAssociationSet", [])
        if cidr["CidrBlockState"]["State"] not in ["disassociating", "disassociated"]
    )

    current_cidrs = set(associated_cidrs.keys())
    desired_cidrs = set(cidr_block)
    if not purge_cidrs:
        desired_cidrs = desired_cidrs.union(current_cidrs)

    cidrs_to_add = list(desired_cidrs.difference(current_cidrs))
    cidrs_to_remove = list(current_cidrs.difference(desired_cidrs))

    if not cidrs_to_add and not cidrs_to_remove:
        return False, None

    if not module.check_mode:
        for cidr in cidrs_to_add:
            try:
                associate_vpc_cidr_block(connection, vpc_id=vpc_obj["VpcId"], CidrBlock=cidr)
            except AnsibleEC2Error as e:
                module.fail_json_aws(e, f"Unable to associate CIDR {cidr}.")

        for cidr in cidrs_to_remove:
            try:
                disassociate_vpc_cidr_block(connection, associated_cidrs[cidr])
            except AnsibleEC2Error as e:
                module.fail_json_aws(
                    e,
                    (
                        f"Unable to disassociate {associated_cidrs[cidr]}. You must detach or delete all gateways and resources"
                        " that are associated with the CIDR block before you can disassociate it."
                    ),
                )
    return True, list(desired_cidrs)


def update_dns_enabled(connection, module, vpc_id, dns_support):
    if dns_support is None:
        return False

    current_dns_enabled = describe_vpc_attribute(connection, attribute="enableDnsSupport", vpc_id=vpc_id)[
        "EnableDnsSupport"
    ]["Value"]
    if current_dns_enabled == dns_support:
        return False

    if module.check_mode:
        return True

    modify_vpc_attribute(connection, vpc_id=vpc_id, EnableDnsSupport={"Value": dns_support})
    return True


def update_dns_hostnames(connection, module, vpc_id, dns_hostnames):
    if dns_hostnames is None:
        return False

    current_dns_hostnames = describe_vpc_attribute(connection, attribute="enableDnsHostnames", vpc_id=vpc_id)[
        "EnableDnsHostnames"
    ]["Value"]
    if current_dns_hostnames == dns_hostnames:
        return False

    if module.check_mode:
        return True

    modify_vpc_attribute(connection, vpc_id=vpc_id, EnableDnsHostnames={"Value": dns_hostnames})
    return True


def wait_for_updates(connection, module, vpc_id, ipv6_cidr, expected_cidrs, dns_support, dns_hostnames, tags, dhcp_id):
    if module.check_mode:
        return

    if expected_cidrs:
        wait_for_resource_state(
            connection,
            module,
            "vpc_available",
            VpcIds=[vpc_id],
            Filters=[{"Name": "cidr-block-association.cidr-block", "Values": expected_cidrs}],
        )
    wait_for_vpc_ipv6_state(module, connection, vpc_id, ipv6_cidr)

    if tags is not None:
        tag_list = ansible_dict_to_boto3_tag_list(tags)
        filters = [{"Name": f"tag:{t['Key']}", "Values": [t["Value"]]} for t in tag_list]
        wait_for_resource_state(connection, module, "vpc_available", VpcIds=[vpc_id], Filters=filters)

    wait_for_vpc_attribute(connection, module, vpc_id, "enableDnsSupport", dns_support)
    wait_for_vpc_attribute(connection, module, vpc_id, "enableDnsHostnames", dns_hostnames)

    if dhcp_id is not None:
        # Wait for DhcpOptionsId to be updated
        filters = [{"Name": "dhcp-options-id", "Values": [dhcp_id]}]
        wait_for_resource_state(connection, module, "vpc_available", VpcIds=[vpc_id], Filters=filters)

    return


def ensure_present(connection, module: AnsibleAWSModule, vpc_id: Optional[str]) -> None:
    name = module.params.get("name")
    cidr_block = module.params.get("cidr_block")
    ipv6_cidr = module.params.get("ipv6_cidr")
    purge_cidrs = module.params.get("purge_cidrs")
    tenancy = module.params.get("tenancy")
    dns_support = module.params.get("dns_support")
    dns_hostnames = module.params.get("dns_hostnames")
    dhcp_id = module.params.get("dhcp_opts_id")
    tags = module.params.get("tags")
    purge_tags = module.params.get("purge_tags")

    if not vpc_id:
        if not name:
            module.fail_json("The name parameter must be specified when creating a new VPC.")
        vpc_obj = create_vpc_net(connection, module, cidr_block[0], tenancy, tags, ipv6_cidr, name)
        vpc_id = vpc_obj["VpcId"]
        changed = True
        # Set on-creation defaults
        if dns_hostnames is None:
            dns_hostnames = True
        if dns_support is None:
            dns_support = True
    else:
        changed = False
        vpc_obj = get_vpc(module, connection, vpc_id)
        changed |= update_ipv6_cidrs(connection, module, vpc_obj, vpc_id, ipv6_cidr)
        changed |= update_vpc_tags(connection, module, vpc_id, tags, name, purge_tags)

    cidrs_changed, desired_cidrs = update_cidrs(connection, module, vpc_obj, cidr_block, purge_cidrs)
    changed |= cidrs_changed
    changed |= update_dhcp_opts(connection, module, vpc_obj, dhcp_id)
    changed |= update_dns_enabled(connection, module, vpc_id, dns_support)
    changed |= update_dns_hostnames(connection, module, vpc_id, dns_hostnames)

    wait_for_updates(connection, module, vpc_id, ipv6_cidr, desired_cidrs, dns_support, dns_hostnames, tags, dhcp_id)

    updated_obj = get_vpc(module, connection, vpc_id)
    final_state = camel_dict_to_snake_dict(updated_obj)
    final_state["tags"] = boto3_tag_list_to_ansible_dict(updated_obj.get("Tags", []))
    final_state["name"] = final_state["tags"].get("Name", None)
    final_state["id"] = final_state.pop("vpc_id")

    module.exit_json(changed=changed, vpc=final_state)


def ensure_absent(connection, module: AnsibleAWSModule, vpc_id: str) -> None:
    changed = False
    if vpc_id:
        if module.check_mode:
            changed = True
        else:
            changed = delete_vpc(connection, vpc_id=vpc_id)
    module.exit_json(changed=changed, vpc={})


def main():
    argument_spec = dict(
        name=dict(required=False),
        vpc_id=dict(type="str", required=False, default=None),
        cidr_block=dict(type="list", elements="str"),
        ipv6_cidr=dict(type="bool", default=None),
        tenancy=dict(choices=["default", "dedicated"], default="default"),
        dns_support=dict(type="bool"),
        dns_hostnames=dict(type="bool"),
        dhcp_opts_id=dict(),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        state=dict(choices=["present", "absent"], default="present"),
        multi_ok=dict(type="bool", default=False),
        purge_cidrs=dict(type="bool", default=False),
    )
    required_one_of = [
        ["vpc_id", "name"],
        ["vpc_id", "cidr_block"],
    ]

    module = AnsibleAWSModule(argument_spec=argument_spec, required_one_of=required_one_of, supports_check_mode=True)

    dns_support = module.params.get("dns_support")
    dns_hostnames = module.params.get("dns_hostnames")
    state = module.params.get("state")

    if dns_hostnames and not dns_support:
        module.fail_json(msg="In order to enable DNS Hostnames you must also enable DNS support")

    connection = module.client("ec2")
    try:
        vpc_id = module.params.get("vpc_id") or get_vpc_id(connection, module)
        if state == "present":
            ensure_present(connection, module, vpc_id)
        else:
            ensure_absent(connection, module, vpc_id)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_anp_epg
short_description: Manage Endpoint Groups (EPGs) in schema templates
description:
- Manage EPGs in schema templates on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
- Anvitha Jain (@anvitha-jain)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  anp:
    description:
    - The name of the ANP.
    type: str
    required: true
  epg:
    description:
    - The name of the EPG to manage.
    type: str
    aliases: [ name ]
  display_name:
    description:
    - The name as displayed on the MSO web interface.
    type: str
  description:
    description:
    - The description as displayed on the MSO web interface.
    - The description is supported on versions of MSO that are 3.3 or greater.
    type: str
#  contracts:
#    description:
#    - A list of contracts associated to this ANP.
#    type: list
  bd:
    description:
    - The BD associated to this ANP.
    type: dict
    suboptions:
      name:
        description:
        - The name of the BD to associate with.
        required: true
        type: str
      schema:
        description:
        - The schema that defines the referenced BD.
        - If this parameter is unspecified, it defaults to the current schema.
        type: str
      template:
        description:
        - The template that defines the referenced BD.
        type: str
  vrf:
    description:
    - The VRF associated to this ANP.
    type: dict
    suboptions:
      name:
        description:
        - The name of the VRF to associate with.
        required: true
        type: str
      schema:
        description:
        - The schema that defines the referenced VRF.
        - If this parameter is unspecified, it defaults to the current schema.
        type: str
      template:
        description:
        - The template that defines the referenced VRF.
        type: str
  subnets:
    description:
    - The subnets associated to this ANP.
    type: list
    elements: dict
    suboptions:
      subnet:
        description:
        - The IP range in CIDR notation.
        type: str
        required: true
        aliases: [ ip ]
      description:
        description:
        - The description of this subnet.
        type: str
      scope:
        description:
        - The scope of the subnet.
        type: str
        default: private
        choices: [ private, public ]
      shared:
        description:
        - Whether this subnet is shared between VRFs.
        type: bool
        default: false
      no_default_gateway:
        description:
        - Whether this subnet has a default gateway.
        type: bool
        default: false
  useg_epg:
    description:
    - Whether this is a USEG EPG.
    type: bool
#  useg_epg_attributes:
#    description:
#    - A dictionary consisting of USEG attributes.
#    type: dict
  intra_epg_isolation:
    description:
    - Whether intra EPG isolation is enforced.
    - When not specified, this parameter defaults to C(unenforced).
    type: str
    choices: [ enforced, unenforced ]
  intersite_multicast_source:
    description:
    - Whether intersite multicast source is enabled.
    - When not specified, this parameter defaults to C(no).
    type: bool
  proxy_arp:
    description:
    - Whether proxy arp is enabled.
    - When not specified, this parameter defaults to C(no).
    type: bool
  preferred_group:
    description:
    - Whether this EPG is added to preferred group or not.
    - When not specified, this parameter defaults to C(no).
    type: bool
  qos_level:
    description:
    - Quality of Service (QoS) allows you to classify the network traffic in the fabric.
    - It helps prioritize and police the traffic flow to help avoid congestion in the network.
    - The Contract QoS Level parameter is supported on versions of MSO that are 3.1 or greater.
    choices: [ unspecified, level1, level2, level3, level4, level5, level6 ]
    type: str
  epg_type:
    description:
    - The EPG type parameter is supported on versions of MSO that are 3.3 or greater.
    type: str
    choices: [ application, service ]
  deployment_type:
    description:
    - The deployment_type parameter indicates how and where the service is deployed.
    - This parameter is available only when epg_type is service.
    type: str
    choices: [ cloud_native, cloud_native_managed, third_party ]
  access_type:
    description:
    - This parameter indicates how the service will be accessed.
    - It is only available when epg_type is service.
    type: str
    choices: [ private, public, public_and_private ]
  service_type:
    description:
    - The service_type parameter refers to the type of cloud services.
    - Only certain deployment types, and certain access types within each deployment type, are supported for each service type.
    - This parameter is available only when epg_type is service.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_template_anp
- module: cisco.mso.mso_schema_template_anp_epg_subnet
- module: cisco.mso.mso_schema_template_bd
- module: cisco.mso.mso_schema_template_contract_filter
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new EPG
  cisco.mso.mso_schema_template_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    bd:
      name: bd1
    vrf:
      name: vrf1
    state: present

- name: Add a new EPG with preferred group.
  cisco.mso.mso_schema_template_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    state: present
    preferred_group: true

- name: Remove an EPG
  cisco.mso.mso_schema_template_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    bd:
      name: bd1
    vrf:
      name: vrf1
    state: absent

- name: Query a specific EPG
  cisco.mso.mso_schema_template_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    bd:
      name: bd1
    vrf:
      name: vrf1
    state: query
  register: query_result

- name: Query all EPGs
  cisco.mso.mso_schema_template_anp_epg:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    anp: ANP 1
    epg: EPG 1
    bd:
      name: bd1
    vrf:
      name: vrf1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec, mso_reference_spec, mso_epg_subnet_spec
from ansible_collections.cisco.mso.plugins.module_utils.constants import QOS_LEVEL


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        anp=dict(type="str", required=True),
        epg=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        description=dict(type="str"),
        bd=dict(type="dict", options=mso_reference_spec()),
        vrf=dict(type="dict", options=mso_reference_spec()),
        display_name=dict(type="str"),
        useg_epg=dict(type="bool"),
        intra_epg_isolation=dict(type="str", choices=["enforced", "unenforced"]),
        intersite_multicast_source=dict(type="bool"),
        proxy_arp=dict(type="bool"),
        subnets=dict(type="list", elements="dict", options=mso_epg_subnet_spec()),
        qos_level=dict(type="str", choices=QOS_LEVEL),
        epg_type=dict(type="str", choices=["application", "service"]),
        deployment_type=dict(type="str", choices=["cloud_native", "cloud_native_managed", "third_party"]),
        service_type=dict(type="str"),
        access_type=dict(type="str", choices=["private", "public", "public_and_private"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        preferred_group=dict(type="bool"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["epg"]],
            ["state", "present", ["epg"]],
        ],
    )

    schema = module.params.get("schema")
    template = module.params.get("template").replace(" ", "")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    description = module.params.get("description")
    display_name = module.params.get("display_name")
    bd = module.params.get("bd")
    if bd is not None and bd.get("template") is not None:
        bd["template"] = bd.get("template").replace(" ", "")
    vrf = module.params.get("vrf")
    if vrf is not None and vrf.get("template") is not None:
        vrf["template"] = vrf.get("template").replace(" ", "")
    useg_epg = module.params.get("useg_epg")
    intra_epg_isolation = module.params.get("intra_epg_isolation")
    intersite_multicast_source = module.params.get("intersite_multicast_source")
    proxy_arp = module.params.get("proxy_arp")
    subnets = module.params.get("subnets")
    qos_level = module.params.get("qos_level")
    epg_type = module.params.get("epg_type")
    deployment_type = module.params.get("deployment_type")
    service_type = module.params.get("service_type")
    access_type = module.params.get("access_type")
    state = module.params.get("state")
    preferred_group = module.params.get("preferred_group")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))
    template_idx = templates.index(template)

    # Get ANP
    anps = [a.get("name") for a in schema_obj.get("templates")[template_idx]["anps"]]
    if anp not in anps:
        mso.fail_json(msg="Provided anp '{0}' does not exist. Existing anps: {1}".format(anp, ", ".join(anps)))
    anp_idx = anps.index(anp)

    # Get EPG
    epgs = [e.get("name") for e in schema_obj.get("templates")[template_idx]["anps"][anp_idx]["epgs"]]
    if epg is not None and epg in epgs:
        epg_idx = epgs.index(epg)
        mso.existing = schema_obj.get("templates")[template_idx]["anps"][anp_idx]["epgs"][epg_idx]

    if state == "query":
        if epg is None:
            mso.existing = schema_obj.get("templates")[template_idx]["anps"][anp_idx]["epgs"]
        elif not mso.existing:
            mso.fail_json(msg="EPG '{epg}' not found".format(epg=epg))

        if "bdRef" in mso.existing:
            mso.existing["bdRef"] = mso.dict_from_ref(mso.existing["bdRef"])
        if "vrfRef" in mso.existing:
            mso.existing["vrfRef"] = mso.dict_from_ref(mso.existing["vrfRef"])
        mso.exit_json()

    epgs_path = "/templates/{0}/anps/{1}/epgs".format(template, anp)
    epg_path = "/templates/{0}/anps/{1}/epgs/{2}".format(template, anp, epg)
    service_path = "{0}/cloudServiceEpgConfig".format(epg_path)
    ops = []
    cloud_service_epg_config = {}

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=epg_path))

    elif state == "present":
        bd_ref = mso.make_reference(bd, "bd", schema_id, template)
        vrf_ref = mso.make_reference(vrf, "vrf", schema_id, template)
        subnets = mso.make_subnets(subnets, is_bd_subnet=False)

        if display_name is None and not mso.existing:
            display_name = epg

        payload = dict(
            name=epg,
            displayName=display_name,
            uSegEpg=useg_epg,
            intraEpg=intra_epg_isolation,
            mCastSource=intersite_multicast_source,
            proxyArp=proxy_arp,
            # FIXME: Missing functionality
            # uSegAttrs=[],
            subnets=subnets,
            bdRef=bd_ref,
            preferredGroup=preferred_group,
            vrfRef=vrf_ref,
        )
        if description is not None:
            payload.update(description=description)
        if qos_level is not None:
            payload.update(prio=qos_level)
        if epg_type is not None:
            payload.update(epgType=epg_type)

        mso.sanitize(payload, collate=True)

        if mso.existing:
            # Clean contractRef to fix api issue
            for contract in mso.sent.get("contractRelationships"):
                contract["contractRef"] = mso.dict_from_ref(contract.get("contractRef"))
            ops.append(dict(op="replace", path=epg_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=epgs_path + "/-", value=mso.sent))

        if epg_type == "service":
            access_type_map = {
                "private": "Private",
                "public": "Public",
                "public_and_private": "PublicAndPrivate",
            }
            deployment_type_map = {
                "cloud_native": "CloudNative",
                "cloud_native_managed": "CloudNativeManaged",
                "third_party": "Third-party",
            }
            if cloud_service_epg_config != {}:
                cloud_service_epg_config.update(
                    dict(deploymentType=deployment_type_map[deployment_type], serviceType=service_type, accessType=access_type_map[access_type])
                )
                ops.append(dict(op="replace", path=service_path, value=cloud_service_epg_config))
            else:
                cloud_service_epg_config.update(
                    dict(deploymentType=deployment_type_map[deployment_type], serviceType=service_type, accessType=access_type_map[access_type])
                )
                ops.append(dict(op="add", path=service_path, value=cloud_service_epg_config))

    mso.existing = mso.proposed

    if "epgRef" in mso.previous:
        del mso.previous["epgRef"]
    if "bdRef" in mso.previous and mso.previous["bdRef"] != "":
        mso.previous["bdRef"] = mso.dict_from_ref(mso.previous["bdRef"])
    if "vrfRef" in mso.previous and mso.previous["bdRef"] != "":
        mso.previous["vrfRef"] = mso.dict_from_ref(mso.previous["vrfRef"])

    if not module.check_mode and mso.proposed != mso.previous:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright 2013 Bruce Pennypacker <bruce@pennypacker.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: airbrake_deployment
author:
  - "Bruce Pennypacker (@bpennypacker)"
  - "Patrick Humpal (@phumpal)"
short_description: Notify airbrake about app deployments
description:
  - Notify airbrake about app deployments (see U(https://airbrake.io/docs/api/#deploys-v4)).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  project_id:
    description:
      - Airbrake PROJECT_ID.
    required: true
    type: str
    version_added: '0.2.0'
  project_key:
    description:
      - Airbrake PROJECT_KEY.
    required: true
    type: str
    version_added: '0.2.0'
  environment:
    description:
      - The airbrake environment name, typically v(production), V(staging), and so on.
    required: true
    type: str
  user:
    description:
      - The username of the person doing the deployment.
    type: str
  repo:
    description:
      - URL of the project repository.
    type: str
  revision:
    description:
      - A hash, number, tag, or other identifier showing what revision from version control was deployed.
    type: str
  version:
    description:
      - A string identifying what version was deployed.
    type: str
    version_added: '1.0.0'
  url:
    description:
      - Optional URL to submit the notification to. Use to send notifications to Airbrake-compliant tools like Errbit.
    default: "https://api.airbrake.io/api/v4/projects/"
    type: str
  validate_certs:
    description:
      - If V(false), SSL certificates for the target URL is not validated. This should only be used on personally controlled
        sites using self-signed certificates.
    default: true
    type: bool

requirements: []
"""

EXAMPLES = r"""
- name: Notify airbrake about an app deployment
  community.general.airbrake_deployment:
    project_id: '12345'
    project_key: 'AAAAAA'
    environment: staging
    user: ansible
    revision: '4.2'

- name: Notify airbrake about an app deployment, using git hash as revision
  community.general.airbrake_deployment:
    project_id: '12345'
    project_key: 'AAAAAA'
    environment: staging
    user: ansible
    revision: 'e54dd3a01f2c421b558ef33b5f79db936e2dcf15'
    version: '0.2.0'
"""

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.urls import fetch_url

# ===========================================
# Module execution.
#


def main():
    module = AnsibleModule(
        argument_spec=dict(
            project_id=dict(required=True, no_log=True, type="str"),
            project_key=dict(required=True, no_log=True, type="str"),
            environment=dict(required=True, type="str"),
            user=dict(type="str"),
            repo=dict(type="str"),
            revision=dict(type="str"),
            version=dict(type="str"),
            url=dict(default="https://api.airbrake.io/api/v4/projects/", type="str"),
            validate_certs=dict(default=True, type="bool"),
        ),
        supports_check_mode=True,
    )

    # Build list of params
    params = {}

    # If we're in check mode, just exit pretending like we succeeded
    if module.check_mode:
        module.exit_json(changed=True)

    # v4 API documented at https://airbrake.io/docs/api/#create-deploy-v4
    if module.params["environment"]:
        params["environment"] = module.params["environment"]

    if module.params["user"]:
        params["username"] = module.params["user"]

    if module.params["repo"]:
        params["repository"] = module.params["repo"]

    if module.params["revision"]:
        params["revision"] = module.params["revision"]

    if module.params["version"]:
        params["version"] = module.params["version"]

    # Build deploy url
    url = f"{module.params.get('url')}{module.params['project_id']}/deploys?key={module.params['project_key']}"
    json_body = module.jsonify(params)

    # Build header
    headers = {"Content-Type": "application/json"}

    # Notify Airbrake of deploy
    response, info = fetch_url(module, url, data=json_body, headers=headers, method="POST")

    if info["status"] == 200 or info["status"] == 201:
        module.exit_json(changed=True)
    else:
        module.fail_json(msg=f"HTTP result code: {info['status']} connecting to {url}")


if __name__ == "__main__":
    main()



#
# Copyright (c) 2018, Bojan Vitnik <bvitnik@mainstream.rs>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import atexit
import re
import time
import traceback
import typing as t

XENAPI_IMP_ERR = None
try:
    import XenAPI

    HAS_XENAPI = True
except ImportError:
    HAS_XENAPI = False
    XENAPI_IMP_ERR = traceback.format_exc()

from ansible.module_utils.ansible_release import __version__ as ANSIBLE_VERSION
from ansible.module_utils.basic import env_fallback, missing_required_lib

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule


def xenserver_common_argument_spec() -> dict[str, t.Any]:
    return dict(
        hostname=dict(
            type="str",
            aliases=["host", "pool"],
            default="localhost",
            fallback=(env_fallback, ["XENSERVER_HOST"]),
        ),
        username=dict(
            type="str", aliases=["user", "admin"], default="root", fallback=(env_fallback, ["XENSERVER_USER"])
        ),
        password=dict(
            type="str", aliases=["pass", "pwd"], no_log=True, fallback=(env_fallback, ["XENSERVER_PASSWORD"])
        ),
        validate_certs=dict(type="bool", default=True, fallback=(env_fallback, ["XENSERVER_VALIDATE_CERTS"])),
    )


def xapi_to_module_vm_power_state(power_state: str) -> str | None:
    """Maps XAPI VM power states to module VM power states."""
    module_power_state_map = {
        "running": "poweredon",
        "halted": "poweredoff",
        "suspended": "suspended",
        "paused": "paused",
    }

    return module_power_state_map.get(power_state)


def module_to_xapi_vm_power_state(power_state: str) -> str | None:
    """Maps module VM power states to XAPI VM power states."""
    vm_power_state_map = {
        "poweredon": "running",
        "poweredoff": "halted",
        "restarted": "running",
        "suspended": "suspended",
        "shutdownguest": "halted",
        "rebootguest": "running",
    }

    return vm_power_state_map.get(power_state)


def is_valid_ip_addr(ip_addr: str) -> bool:
    """Validates given string as IPv4 address for given string.

    Args:
        ip_addr (str): string to validate as IPv4 address.

    Returns:
        bool: True if string is valid IPv4 address, else False.
    """
    ip_addr_split = ip_addr.split(".")

    if len(ip_addr_split) != 4:
        return False

    for ip_addr_octet in ip_addr_split:
        if not ip_addr_octet.isdigit():
            return False

        ip_addr_octet_int = int(ip_addr_octet)

        if ip_addr_octet_int < 0 or ip_addr_octet_int > 255:
            return False

    return True


def is_valid_ip_netmask(ip_netmask: str) -> bool:
    """Validates given string as IPv4 netmask.

    Args:
        ip_netmask (str): string to validate as IPv4 netmask.

    Returns:
        bool: True if string is valid IPv4 netmask, else False.
    """
    ip_netmask_split = ip_netmask.split(".")

    if len(ip_netmask_split) != 4:
        return False

    valid_octet_values = ["0", "128", "192", "224", "240", "248", "252", "254", "255"]

    for ip_netmask_octet in ip_netmask_split:
        if ip_netmask_octet not in valid_octet_values:
            return False

    if ip_netmask_split[0] != "255" and (
        ip_netmask_split[1] != "0" or ip_netmask_split[2] != "0" or ip_netmask_split[3] != "0"
    ):
        return False
    elif ip_netmask_split[1] != "255" and (ip_netmask_split[2] != "0" or ip_netmask_split[3] != "0"):
        return False
    elif ip_netmask_split[2] != "255" and ip_netmask_split[3] != "0":
        return False

    return True


def is_valid_ip_prefix(ip_prefix: str) -> bool:
    """Validates given string as IPv4 prefix.

    Args:
        ip_prefix (str): string to validate as IPv4 prefix.

    Returns:
        bool: True if string is valid IPv4 prefix, else False.
    """
    if not ip_prefix.isdigit():
        return False

    ip_prefix_int = int(ip_prefix)

    return not (ip_prefix_int < 0 or ip_prefix_int > 32)


def ip_prefix_to_netmask(ip_prefix: str, skip_check: bool = False) -> str:
    """Converts IPv4 prefix to netmask.

    Args:
        ip_prefix (str): IPv4 prefix to convert.
        skip_check (bool): Skip validation of IPv4 prefix
            (default: False). Use if you are sure IPv4 prefix is valid.

    Returns:
        str: IPv4 netmask equivalent to given IPv4 prefix if
        IPv4 prefix is valid, else an empty string.
    """
    if skip_check:
        ip_prefix_valid = True
    else:
        ip_prefix_valid = is_valid_ip_prefix(ip_prefix)

    if ip_prefix_valid:
        return ".".join([str((0xFFFFFFFF << (32 - int(ip_prefix)) >> i) & 0xFF) for i in [24, 16, 8, 0]])
    else:
        return ""


def ip_netmask_to_prefix(ip_netmask: str, skip_check: bool = False) -> str:
    """Converts IPv4 netmask to prefix.

    Args:
        ip_netmask (str): IPv4 netmask to convert.
        skip_check (bool): Skip validation of IPv4 netmask
            (default: False). Use if you are sure IPv4 netmask is valid.

    Returns:
        str: IPv4 prefix equivalent to given IPv4 netmask if
        IPv4 netmask is valid, else an empty string.
    """
    if skip_check:
        ip_netmask_valid = True
    else:
        ip_netmask_valid = is_valid_ip_netmask(ip_netmask)

    if ip_netmask_valid:
        return str(sum([bin(int(i)).count("1") for i in ip_netmask.split(".")]))
    else:
        return ""


def is_valid_ip6_addr(ip6_addr: str) -> bool:
    """Validates given string as IPv6 address.

    Args:
        ip6_addr (str): string to validate as IPv6 address.

    Returns:
        bool: True if string is valid IPv6 address, else False.
    """
    ip6_addr = ip6_addr.lower()
    ip6_addr_split = ip6_addr.split(":")

    if ip6_addr_split[0] == "":
        ip6_addr_split.pop(0)

    if ip6_addr_split[-1] == "":
        ip6_addr_split.pop(-1)

    if len(ip6_addr_split) > 8:
        return False

    if ip6_addr_split.count("") > 1:
        return False
    elif ip6_addr_split.count("") == 1:
        ip6_addr_split.remove("")
    else:
        if len(ip6_addr_split) != 8:
            return False

    ip6_addr_hextet_regex = re.compile("^[0-9a-f]{1,4}$")

    return all(ip6_addr_hextet_regex.match(ip6_addr_hextet) for ip6_addr_hextet in ip6_addr_split)


def is_valid_ip6_prefix(ip6_prefix: str) -> bool:
    """Validates given string as IPv6 prefix.

    Args:
        ip6_prefix (str): string to validate as IPv6 prefix.

    Returns:
        bool: True if string is valid IPv6 prefix, else False.
    """
    if not ip6_prefix.isdigit():
        return False

    ip6_prefix_int = int(ip6_prefix)

    return not (ip6_prefix_int < 0 or ip6_prefix_int > 128)


def get_object_ref(module: AnsibleModule, name, uuid=None, obj_type="VM", fail=True, msg_prefix=""):
    """Finds and returns a reference to arbitrary XAPI object.

    An object is searched by using either name (name_label) or UUID
    with UUID taken precedence over name.

    Args:
        module: Reference to Ansible module object.
        name (str): Name (name_label) of an object to search for.
        uuid (str): UUID of an object to search for.
        obj_type (str): Any valid XAPI object type. See XAPI docs.
        fail (bool): Should function fail with error message if object
            is not found or exit silently (default: True). The function
            always fails if multiple objects with same name are found.
        msg_prefix (str): A string error messages should be prefixed
            with (default: "").

    Returns:
        XAPI reference to found object or None if object is not found
        and fail=False.
    """
    xapi_session = XAPI.connect(module)

    if obj_type in ["template", "snapshot"]:
        real_obj_type = "VM"
    elif obj_type == "home server":
        real_obj_type = "host"
    elif obj_type == "ISO image":
        real_obj_type = "VDI"
    else:
        real_obj_type = obj_type

    obj_ref = None

    # UUID has precedence over name.
    if uuid:
        try:
            # Find object by UUID. If no object is found using given UUID,
            # an exception will be generated.
            obj_ref = xapi_session.xenapi_request(f"{real_obj_type}.get_by_uuid", (uuid,))
        except XenAPI.Failure:
            if fail:
                module.fail_json(msg=f"{msg_prefix}{obj_type} with UUID '{uuid}' not found!")
    elif name:
        try:
            # Find object by name (name_label).
            obj_ref_list = xapi_session.xenapi_request(f"{real_obj_type}.get_by_name_label", (name,))
        except XenAPI.Failure as f:
            module.fail_json(msg=f"XAPI ERROR: {f.details}")

        # If obj_ref_list is empty.
        if not obj_ref_list:
            if fail:
                module.fail_json(msg=f"{msg_prefix}{obj_type} with name '{name}' not found!")
        # If obj_ref_list contains multiple object references.
        elif len(obj_ref_list) > 1:
            module.fail_json(msg=f"{msg_prefix}multiple {obj_type}s with name '{name}' found! Please use UUID.")
        # The obj_ref_list contains only one object reference.
        else:
            obj_ref = obj_ref_list[0]
    else:
        module.fail_json(msg=f"{msg_prefix}no valid name or UUID supplied for {obj_type}!")

    return obj_ref


def gather_vm_params(module: AnsibleModule, vm_ref):
    """Gathers all VM parameters available in XAPI database.

    Args:
        module: Reference to Ansible module object.
        vm_ref (str): XAPI reference to VM.

    Returns:
        dict: VM parameters.
    """
    # We silently return empty vm_params if bad vm_ref was supplied.
    if not vm_ref or vm_ref == "OpaqueRef:NULL":
        return {}

    xapi_session = XAPI.connect(module)

    try:
        vm_params = xapi_session.xenapi.VM.get_record(vm_ref)

        # We need some params like affinity, VBDs, VIFs, VDIs etc. dereferenced.

        # Affinity.
        if vm_params["affinity"] != "OpaqueRef:NULL":
            vm_affinity = xapi_session.xenapi.host.get_record(vm_params["affinity"])
            vm_params["affinity"] = vm_affinity
        else:
            vm_params["affinity"] = {}

        # VBDs.
        vm_vbd_params_list = [xapi_session.xenapi.VBD.get_record(vm_vbd_ref) for vm_vbd_ref in vm_params["VBDs"]]

        # List of VBDs is usually sorted by userdevice but we sort just
        # in case. We need this list sorted by userdevice so that we can
        # make positional pairing with module.params['disks'].
        vm_vbd_params_list = sorted(vm_vbd_params_list, key=lambda vm_vbd_params: int(vm_vbd_params["userdevice"]))
        vm_params["VBDs"] = vm_vbd_params_list

        # VDIs.
        for vm_vbd_params in vm_params["VBDs"]:
            if vm_vbd_params["VDI"] != "OpaqueRef:NULL":
                vm_vdi_params = xapi_session.xenapi.VDI.get_record(vm_vbd_params["VDI"])
            else:
                vm_vdi_params = {}

            vm_vbd_params["VDI"] = vm_vdi_params

        # VIFs.
        vm_vif_params_list = [xapi_session.xenapi.VIF.get_record(vm_vif_ref) for vm_vif_ref in vm_params["VIFs"]]

        # List of VIFs is usually sorted by device but we sort just
        # in case. We need this list sorted by device so that we can
        # make positional pairing with module.params['networks'].
        vm_vif_params_list = sorted(vm_vif_params_list, key=lambda vm_vif_params: int(vm_vif_params["device"]))
        vm_params["VIFs"] = vm_vif_params_list

        # Networks.
        for vm_vif_params in vm_params["VIFs"]:
            if vm_vif_params["network"] != "OpaqueRef:NULL":
                vm_network_params = xapi_session.xenapi.network.get_record(vm_vif_params["network"])
            else:
                vm_network_params = {}

            vm_vif_params["network"] = vm_network_params

        # Guest metrics.
        if vm_params["guest_metrics"] != "OpaqueRef:NULL":
            vm_guest_metrics = xapi_session.xenapi.VM_guest_metrics.get_record(vm_params["guest_metrics"])
            vm_params["guest_metrics"] = vm_guest_metrics
        else:
            vm_params["guest_metrics"] = {}

        # Detect customization agent.
        xenserver_version = get_xenserver_version(module)

        if (
            xenserver_version[0] >= 7
            and xenserver_version[1] >= 0
            and vm_params.get("guest_metrics")
            and "feature-static-ip-setting" in vm_params["guest_metrics"]["other"]
        ):
            vm_params["customization_agent"] = "native"
        else:
            vm_params["customization_agent"] = "custom"

    except XenAPI.Failure as f:
        module.fail_json(msg=f"XAPI ERROR: {f.details}")

    return vm_params


def gather_vm_facts(module: AnsibleModule, vm_params):
    """Gathers VM facts.

    Args:
        module: Reference to Ansible module object.
        vm_params (dict): A dictionary with VM parameters as returned
            by gather_vm_params() function.

    Returns:
        dict: VM facts.
    """
    # We silently return empty vm_facts if no vm_params are available.
    if not vm_params:
        return {}

    xapi_session = XAPI.connect(module)

    # Gather facts.
    vm_facts = {
        "state": xapi_to_module_vm_power_state(vm_params["power_state"].lower()),
        "name": vm_params["name_label"],
        "name_desc": vm_params["name_description"],
        "uuid": vm_params["uuid"],
        "is_template": vm_params["is_a_template"],
        "folder": vm_params["other_config"].get("folder", ""),
        "hardware": {
            "num_cpus": int(vm_params["VCPUs_max"]),
            "num_cpu_cores_per_socket": int(vm_params["platform"].get("cores-per-socket", "1")),
            "memory_mb": int(int(vm_params["memory_dynamic_max"]) / 1048576),
        },
        "disks": [],
        "cdrom": {},
        "networks": [],
        "home_server": vm_params["affinity"].get("name_label", ""),
        "domid": vm_params["domid"],
        "platform": vm_params["platform"],
        "other_config": vm_params["other_config"],
        "xenstore_data": vm_params["xenstore_data"],
        "customization_agent": vm_params["customization_agent"],
    }

    for vm_vbd_params in vm_params["VBDs"]:
        if vm_vbd_params["type"] == "Disk":
            vm_disk_sr_params = xapi_session.xenapi.SR.get_record(vm_vbd_params["VDI"]["SR"])

            vm_disk_params = {
                "size": int(vm_vbd_params["VDI"]["virtual_size"]),
                "name": vm_vbd_params["VDI"]["name_label"],
                "name_desc": vm_vbd_params["VDI"]["name_description"],
                "sr": vm_disk_sr_params["name_label"],
                "sr_uuid": vm_disk_sr_params["uuid"],
                "os_device": vm_vbd_params["device"],
                "uuid": vm_vbd_params["VDI"]["uuid"],
                "vbd_userdevice": vm_vbd_params["userdevice"],
                "vdi_type": vm_vbd_params["VDI"]["sm_config"].get(
                    "image-format",
                    vm_vbd_params["VDI"]["sm_config"].get(
                        "vdi_type",
                        vm_vbd_params["VDI"]["sm_config"].get("type", "raw"),
                    ),
                ),
            }

            vm_facts["disks"].append(vm_disk_params)
        elif vm_vbd_params["type"] == "CD":
            if vm_vbd_params["empty"]:
                vm_facts["cdrom"].update(type="none")
            else:
                vm_facts["cdrom"].update(type="iso")
                vm_facts["cdrom"].update(iso_name=vm_vbd_params["VDI"]["name_label"])

    for vm_vif_params in vm_params["VIFs"]:
        vm_guest_metrics_networks = vm_params["guest_metrics"].get("networks", {})

        vm_network_params = {
            "name": vm_vif_params["network"]["name_label"],
            "mac": vm_vif_params["MAC"],
            "vif_device": vm_vif_params["device"],
            "mtu": vm_vif_params["MTU"],
            "ip": vm_guest_metrics_networks.get(f"{vm_vif_params['device']}/ip", ""),
            "prefix": "",
            "netmask": "",
            "gateway": "",
            "ip6": [
                vm_guest_metrics_networks[ipv6]
                for ipv6 in sorted(vm_guest_metrics_networks.keys())
                if ipv6.startswith(f"{vm_vif_params['device']}/ipv6/")
            ],
            "prefix6": "",
            "gateway6": "",
        }

        if vm_params["customization_agent"] == "native":
            if vm_vif_params["ipv4_addresses"] and vm_vif_params["ipv4_addresses"][0]:
                vm_network_params["prefix"] = vm_vif_params["ipv4_addresses"][0].split("/")[1]
                vm_network_params["netmask"] = ip_prefix_to_netmask(vm_network_params["prefix"])

            vm_network_params["gateway"] = vm_vif_params["ipv4_gateway"]

            if vm_vif_params["ipv6_addresses"] and vm_vif_params["ipv6_addresses"][0]:
                vm_network_params["prefix6"] = vm_vif_params["ipv6_addresses"][0].split("/")[1]

            vm_network_params["gateway6"] = vm_vif_params["ipv6_gateway"]

        elif vm_params["customization_agent"] == "custom":
            vm_xenstore_data = vm_params["xenstore_data"]

            for f in ["prefix", "netmask", "gateway", "prefix6", "gateway6"]:
                vm_network_params[f] = vm_xenstore_data.get(f"vm-data/networks/{vm_vif_params['device']}/{f}", "")

        vm_facts["networks"].append(vm_network_params)

    return vm_facts


def set_vm_power_state(module: AnsibleModule, vm_ref, power_state, timeout=300):
    """Controls VM power state.

    Args:
        module: Reference to Ansible module object.
        vm_ref (str): XAPI reference to VM.
        power_state (str): Power state to put VM into. Accepted values:

            - poweredon
            - poweredoff
            - restarted
            - suspended
            - shutdownguest
            - rebootguest

        timeout (int): timeout in seconds (default: 300).

    Returns:
        tuple (bool, str): Bool element is True if VM power state has
        changed by calling this function, else False. Str element carries
        a value of resulting power state as defined by XAPI - 'running',
        'halted' or 'suspended'.
    """
    # Fail if we don't have a valid VM reference.
    if not vm_ref or vm_ref == "OpaqueRef:NULL":
        module.fail_json(msg="Cannot set VM power state. Invalid VM reference supplied!")

    xapi_session = XAPI.connect(module)

    power_state = power_state.replace("_", "").replace("-", "").lower()
    vm_power_state_resulting = module_to_xapi_vm_power_state(power_state)

    state_changed = False

    try:
        # Get current state of the VM.
        vm_power_state_current = xapi_to_module_vm_power_state(xapi_session.xenapi.VM.get_power_state(vm_ref).lower())

        if vm_power_state_current != power_state:
            if power_state == "poweredon":
                if not module.check_mode:
                    # VM can be in either halted, suspended, paused or running state.
                    # For VM to be in running state, start has to be called on halted,
                    # resume on suspended and unpause on paused VM.
                    if vm_power_state_current == "poweredoff":
                        xapi_session.xenapi.VM.start(vm_ref, False, False)
                    elif vm_power_state_current == "suspended":
                        xapi_session.xenapi.VM.resume(vm_ref, False, False)
                    elif vm_power_state_current == "paused":
                        xapi_session.xenapi.VM.unpause(vm_ref)
            elif power_state == "poweredoff":
                if not module.check_mode:
                    # hard_shutdown will halt VM regardless of current state.
                    xapi_session.xenapi.VM.hard_shutdown(vm_ref)
            elif power_state == "restarted":
                # hard_reboot will restart VM only if VM is in paused or running state.
                if vm_power_state_current in ["paused", "poweredon"]:
                    if not module.check_mode:
                        xapi_session.xenapi.VM.hard_reboot(vm_ref)
                else:
                    module.fail_json(msg=f"Cannot restart VM in state '{vm_power_state_current}'!")
            elif power_state == "suspended":
                # running state is required for suspend.
                if vm_power_state_current == "poweredon":
                    if not module.check_mode:
                        xapi_session.xenapi.VM.suspend(vm_ref)
                else:
                    module.fail_json(msg=f"Cannot suspend VM in state '{vm_power_state_current}'!")
            elif power_state == "shutdownguest":
                # running state is required for guest shutdown.
                if vm_power_state_current == "poweredon":
                    if not module.check_mode:
                        if timeout == 0:
                            xapi_session.xenapi.VM.clean_shutdown(vm_ref)
                        else:
                            task_ref = xapi_session.xenapi.Async.VM.clean_shutdown(vm_ref)
                            task_result = wait_for_task(module, task_ref, timeout)

                            if task_result:
                                module.fail_json(msg=f"Guest shutdown task failed: '{task_result}'!")
                else:
                    module.fail_json(msg=f"Cannot shutdown guest when VM is in state '{vm_power_state_current}'!")
            elif power_state == "rebootguest":
                # running state is required for guest reboot.
                if vm_power_state_current == "poweredon":
                    if not module.check_mode:
                        if timeout == 0:
                            xapi_session.xenapi.VM.clean_reboot(vm_ref)
                        else:
                            task_ref = xapi_session.xenapi.Async.VM.clean_reboot(vm_ref)
                            task_result = wait_for_task(module, task_ref, timeout)

                            if task_result:
                                module.fail_json(msg=f"Guest reboot task failed: '{task_result}'!")
                else:
                    module.fail_json(msg=f"Cannot reboot guest when VM is in state '{vm_power_state_current}'!")
            else:
                module.fail_json(msg=f"Requested VM power state '{power_state}' is unsupported!")

            state_changed = True
    except XenAPI.Failure as f:
        module.fail_json(msg=f"XAPI ERROR: {f.details}")

    return (state_changed, vm_power_state_resulting)


def wait_for_task(module: AnsibleModule, task_ref, timeout=300):
    """Waits for async XAPI task to finish.

    Args:
        module: Reference to Ansible module object.
        task_ref (str): XAPI reference to task.
        timeout (int): timeout in seconds (default: 300).

    Returns:
        str: failure message on failure, else an empty string.
    """
    # Fail if we don't have a valid task reference.
    if not task_ref or task_ref == "OpaqueRef:NULL":
        module.fail_json(msg="Cannot wait for task. Invalid task reference supplied!")

    xapi_session = XAPI.connect(module)

    interval = 2

    result = ""

    # If we have to wait indefinitely, make time_left larger than 0 so we can
    # enter while loop.
    if timeout == 0:
        time_left = 1
    else:
        time_left = timeout

    try:
        while time_left > 0:
            task_status = xapi_session.xenapi.task.get_status(task_ref).lower()

            if task_status == "pending":
                # Task is still running.
                time.sleep(interval)

                # We decrease time_left only if we don't wait indefinitely.
                if timeout != 0:
                    time_left -= interval

                continue
            elif task_status == "success":
                # Task is done.
                break
            else:
                # Task failed.
                result = task_status
                break
        else:
            # We timed out.
            result = "timeout"

        xapi_session.xenapi.task.destroy(task_ref)
    except XenAPI.Failure as f:
        module.fail_json(msg=f"XAPI ERROR: {f.details}")

    return result


def wait_for_vm_ip_address(module: AnsibleModule, vm_ref, timeout=300):
    """Waits for VM to acquire an IP address.

    Args:
        module: Reference to Ansible module object.
        vm_ref (str): XAPI reference to VM.
        timeout (int): timeout in seconds (default: 300).

    Returns:
        dict: VM guest metrics as retrieved by
        VM_guest_metrics.get_record() XAPI method with info
        on IP address acquired.
    """
    # Fail if we don't have a valid VM reference.
    if not vm_ref or vm_ref == "OpaqueRef:NULL":
        module.fail_json(msg="Cannot wait for VM IP address. Invalid VM reference supplied!")

    xapi_session = XAPI.connect(module)

    vm_guest_metrics = {}

    try:
        # We translate VM power state string so that error message can be
        # consistent with module VM power states.
        vm_power_state = xapi_to_module_vm_power_state(xapi_session.xenapi.VM.get_power_state(vm_ref).lower())

        if vm_power_state != "poweredon":
            module.fail_json(msg=f"Cannot wait for VM IP address when VM is in state '{vm_power_state}'!")

        interval = 2

        # If we have to wait indefinitely, make time_left larger than 0 so we can
        # enter while loop.
        if timeout == 0:
            time_left = 1
        else:
            time_left = timeout

        while time_left > 0:
            vm_guest_metrics_ref = xapi_session.xenapi.VM.get_guest_metrics(vm_ref)

            if vm_guest_metrics_ref != "OpaqueRef:NULL":
                vm_guest_metrics = xapi_session.xenapi.VM_guest_metrics.get_record(vm_guest_metrics_ref)
                vm_ips = vm_guest_metrics["networks"]

                if "0/ip" in vm_ips:
                    break

            time.sleep(interval)

            # We decrease time_left only if we don't wait indefinitely.
            if timeout != 0:
                time_left -= interval
        else:
            # We timed out.
            module.fail_json(msg="Timed out waiting for VM IP address!")

    except XenAPI.Failure as f:
        module.fail_json(msg=f"XAPI ERROR: {f.details}")

    return vm_guest_metrics


def get_xenserver_version(module: AnsibleModule):
    """Returns XenServer version.

    Args:
        module: Reference to Ansible module object.

    Returns:
        list: Element [0] is major version. Element [1] is minor version.
        Element [2] is update number.
    """
    xapi_session = XAPI.connect(module)

    host_ref = xapi_session.xenapi.session.get_this_host(xapi_session._session)

    try:
        xenserver_version = [
            int(version_number)
            for version_number in xapi_session.xenapi.host.get_software_version(host_ref)["product_version"].split(".")
        ]
    except ValueError:
        xenserver_version = [0, 0, 0]

    return xenserver_version


class XAPI:
    """Class for XAPI session management."""

    _xapi_session: t.Any | None = None

    @classmethod
    def connect(cls, module: AnsibleModule, disconnect_atexit=True):
        """Establishes XAPI connection and returns session reference.

        If no existing session is available, establishes a new one
        and returns it, else returns existing one.

        Args:
            module: Reference to Ansible module object.
            disconnect_atexit (bool): Controls if method should
                register atexit handler to disconnect from XenServer
                on module exit (default: True).

        Returns:
            XAPI session reference.
        """
        if cls._xapi_session is not None:
            return cls._xapi_session

        hostname = module.params["hostname"]
        username = module.params["username"]
        password = module.params["password"]
        ignore_ssl = not module.params["validate_certs"]

        if hostname == "localhost":
            cls._xapi_session = XenAPI.xapi_local()
            username = ""
            password = ""
        else:
            # If scheme is not specified we default to http:// because https://
            # is problematic in most setups.
            if not hostname.startswith("http://") and not hostname.startswith("https://"):
                hostname = f"http://{hostname}"

            try:
                # ignore_ssl is supported in XenAPI library from XenServer 7.2
                # SDK onward but there is no way to tell which version we
                # are using. TypeError will be raised if ignore_ssl is not
                # supported. Additionally, ignore_ssl requires Python 2.7.9
                # or newer.
                cls._xapi_session = XenAPI.Session(hostname, ignore_ssl=ignore_ssl)
            except TypeError:
                # Try without ignore_ssl.
                cls._xapi_session = XenAPI.Session(hostname)

            if not password:
                password = ""

        try:
            cls._xapi_session.login_with_password(username, password, ANSIBLE_VERSION, "Ansible")
        except XenAPI.Failure as f:
            module.fail_json(msg=f"Unable to log on to XenServer at {hostname} as {username}: {f.details}")

        # Disabling atexit should be used in special cases only.
        if disconnect_atexit:
            atexit.register(cls._xapi_session.logout)

        return cls._xapi_session


class XenServerObject:
    """Base class for all XenServer objects.

    This class contains active XAPI session reference and common
    attributes with useful info about XenServer host/pool.

    Attributes:
        module: Reference to Ansible module object.
        xapi_session: Reference to XAPI session.
        pool_ref (str): XAPI reference to a pool currently connected to.
        default_sr_ref (str): XAPI reference to a pool default
            Storage Repository.
        host_ref (str): XAPI rerefence to a host currently connected to.
        xenserver_version (list of str): Contains XenServer major and
            minor version.
    """

    def __init__(self, module: AnsibleModule) -> None:
        """Inits XenServerObject using common module parameters.

        Args:
            module: Reference to Ansible module object.
        """
        if not HAS_XENAPI:
            module.fail_json(changed=False, msg=missing_required_lib("XenAPI"), exception=XENAPI_IMP_ERR)

        self.module = module
        self.xapi_session = XAPI.connect(module)

        try:
            self.pool_ref = self.xapi_session.xenapi.pool.get_all()[0]
            self.default_sr_ref = self.xapi_session.xenapi.pool.get_default_SR(self.pool_ref)
            self.xenserver_version = get_xenserver_version(module)
        except XenAPI.Failure as f:
            self.module.fail_json(msg=f"XAPI ERROR: {f.details}")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ec2_vpc_subnet
version_added: 1.0.0
short_description: Manage subnets in AWS virtual private clouds
description:
  - Manage subnets in AWS virtual private clouds.
author:
  - Robert Estelle (@erydo)
  - Brad Davidson (@brandond)
options:
  az:
    description:
      - The availability zone for the subnet.
      - Required if O(outpost_arn) is set.
    type: str
  cidr:
    description:
      - The CIDR block for the subnet. E.g. V(192.0.2.0/24).
    type: str
    required: true
  ipv6_cidr:
    description:
      - The IPv6 CIDR block for the subnet.
      - The VPC must have a /56 block assigned and this value must be a valid IPv6 /64 that falls in the VPC range.
      - Required if O(assign_instances_ipv6=true)
    type: str
    default: ''
  outpost_arn:
    description:
      - The Amazon Resource Name (ARN) of the Outpost.
      - If set, allows to create subnet in an Outpost.
      - If O(outpost_arn) is set, O(az) must also be specified.
    type: str
    default: ''
  state:
    description:
      - Create or remove the subnet.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  vpc_id:
    description:
      -"VPC ID of the VPC in which to create or delete the subnet.
    required: true
    type: str
  map_public:
    description:
      - Whether instances launched into the subnet should default to being assigned public IP address.
    type: bool
    default: false
  assign_instances_ipv6:
    description:
      - Whether instances launched into the subnet should default to being automatically assigned an IPv6 address.
      - If O(assign_instances_ipv6=true), O(ipv6_cidr) must also be specified.
    type: bool
    default: false
  wait:
    description:
      - Whether to wait for changes to complete.
    type: bool
    default: true
  wait_timeout:
    description:
      - Number of seconds to wait for changes to complete
      - Ignored unless O(wait=true).
    default: 300
    type: int
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Create subnet for database servers
  amazon.aws.ec2_vpc_subnet:
    state: present
    vpc_id: vpc-123456
    cidr: 10.0.1.16/28
    tags:
      Name: Database Subnet
  register: database_subnet

- name: Remove subnet for database servers
  amazon.aws.ec2_vpc_subnet:
    state: absent
    vpc_id: vpc-123456
    cidr: 10.0.1.16/28

- name: Create subnet with IPv6 block assigned
  amazon.aws.ec2_vpc_subnet:
    state: present
    vpc_id: vpc-123456
    cidr: 10.1.100.0/24
    ipv6_cidr: 2001:db8:0:102::/64

- name: Remove IPv6 block assigned to subnet
  amazon.aws.ec2_vpc_subnet:
    state: present
    vpc_id: vpc-123456
    cidr: 10.1.100.0/24
    ipv6_cidr: ''
"""

RETURN = r"""
subnet:
    description: Dictionary of subnet values.
    returned: O(state=present)
    type: complex
    contains:
        id:
            description: Subnet resource id.
            returned: O(state=present)
            type: str
            sample: subnet-b883b2c4
        cidr_block:
            description: The IPv4 CIDR of the Subnet.
            returned: O(state=present)
            type: str
            sample: "10.0.0.0/16"
        ipv6_cidr_block:
            description: The IPv6 CIDR block actively associated with the Subnet.
            returned: O(state=present)
            type: str
            sample: "2001:db8:0:102::/64"
        availability_zone:
            description: Availability zone of the Subnet.
            returned: O(state=present)
            type: str
            sample: us-east-1a
        availability_zone_id:
            description: The AZ ID of the subnet.
            returned: O(state=present)
            type: str
            sample: use1-az6
        state:
            description: State of the Subnet.
            returned: O(state=present)
            type: str
            sample: available
        tags:
            description: Tags attached to the Subnet, includes name.
            returned: O(state=present)
            type: dict
            sample: {"Name": "My Subnet", "env": "staging"}
        map_public_ip_on_launch:
            description: Whether public IP is auto-assigned to new instances.
            returned: O(state=present)
            type: bool
            sample: false
        assign_ipv6_address_on_creation:
            description: Whether IPv6 address is auto-assigned to new instances.
            returned: O(state=present)
            type: bool
            sample: false
        vpc_id:
            description: The id of the VPC where this Subnet exists.
            returned: O(state=present)
            type: str
            sample: vpc-67236184
        available_ip_address_count:
            description: Number of available IPv4 addresses.
            returned: O(state=present)
            type: str
            sample: 251
        default_for_az:
            description: Indicates whether this is the default Subnet for this Availability Zone.
            returned: O(state=present)
            type: bool
            sample: false
        enable_dns64:
            description:
            - Indicates whether DNS queries made should return synthetic IPv6 addresses for IPv4-only destinations.
            type: bool
            sample: false
        ipv6_association_id:
            description: The IPv6 association ID for the currently associated CIDR.
            returned: O(state=present)
            type: str
            sample: subnet-cidr-assoc-b85c74d2
        ipv6_native:
            description: Indicates whether this is an IPv6 only subnet.
            type: bool
            sample: false
        ipv6_cidr_block_association_set:
            description: An array of IPv6 cidr block association set information.
            returned: O(state=present)
            type: complex
            contains:
                association_id:
                    description: The association ID.
                    returned: always
                    type: str
                ipv6_cidr_block:
                    description: The IPv6 CIDR block that is associated with the subnet.
                    returned: always
                    type: str
                ipv6_cidr_block_state:
                    description: A hash/dict that contains a single item. The state of the cidr block association.
                    returned: always
                    type: dict
                    contains:
                        state:
                            description: The CIDR block association state.
                            returned: always
                            type: str
        map_customer_owned_ip_on_launch:
            description:
            - Indicates whether a network interface receives a customer-owned IPv4 address.
            type: bool
            sample: flase
        owner_id:
            description: The ID of the Amazon Web Services account that owns the subnet.
            type: str
            sample: 12344567
        private_dns_name_options_on_launch:
            description:
            - The type of hostnames to assign to instances in the subnet at launch.
            - An instance hostname is based on the IPv4 address or ID of the instance.
            type: dict
            sample: {
                "enable_resource_name_dns_a_record": false,
                "enable_resource_name_dns_aaaa_record": false,
                "hostname_type": "ip-name"
            }
        subnet_arn:
            description: The Amazon Resource Name (ARN) of the subnet.
            type: str
            sample: arn:aws:ec2:us-east-1:xxx:subnet/subnet-xxx

"""


import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.arn import is_outpost_arn
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AnsibleEC2Error
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import associate_subnet_cidr_block
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import create_subnet
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import delete_subnet
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import describe_subnets
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import disassociate_subnet_cidr_block
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import ensure_ec2_tags
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import modify_subnet_attribute
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_tag_filter_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_specifications
from ansible_collections.amazon.aws.plugins.module_utils.transformation import ansible_dict_to_boto3_filter_list
from ansible_collections.amazon.aws.plugins.module_utils.waiters import wait_for_resource_state


def get_subnet_info(subnet: Dict[str, Any]) -> Dict[str, Any]:
    subnet = camel_dict_to_snake_dict(subnet)
    if "tags" in subnet:
        subnet["tags"] = boto3_tag_list_to_ansible_dict(subnet["tags"])
    else:
        subnet["tags"] = dict()

    if "subnet_id" in subnet:
        subnet["id"] = subnet["subnet_id"]
        del subnet["subnet_id"]

    subnet["ipv6_cidr_block"] = ""
    subnet["ipv6_association_id"] = ""
    ipv6set = subnet.get("ipv6_cidr_block_association_set")
    if ipv6set:
        for item in ipv6set:
            if item.get("ipv6_cidr_block_state", {}).get("state") in ("associated", "associating"):
                subnet["ipv6_cidr_block"] = item["ipv6_cidr_block"]
                subnet["ipv6_association_id"] = item["association_id"]

    return subnet


def wait_vpc_subnet_state(conn, module: AnsibleAWSModule, waiter_name: str, start_time: float, **params) -> None:
    max_attempts = int(module.params["wait_timeout"] + start_time - time.time())
    wait_for_resource_state(conn, module, waiter_name, delay=5, max_attempts=max_attempts, **params)


def create_vpc_subnet(
    connection,
    module: AnsibleAWSModule,
    start_time: float,
) -> Dict[str, Any]:
    vpc_id = module.params["vpc_id"]
    cidr = module.params["cidr"]
    tags = module.params["tags"]
    ipv6_cidr = module.params["ipv6_cidr"]
    outpost_arn = module.params["outpost_arn"]
    az = module.params["az"]
    wait = module.params["wait"]

    params = {"VpcId": vpc_id, "CidrBlock": cidr}

    if ipv6_cidr:
        params["Ipv6CidrBlock"] = ipv6_cidr
    if az:
        params["AvailabilityZone"] = az
    if tags:
        params["TagSpecifications"] = boto3_tag_specifications(tags, types="subnet")
    if outpost_arn:
        if is_outpost_arn(outpost_arn):
            params["OutpostArn"] = outpost_arn
        else:
            module.fail_json("OutpostArn does not match the pattern specified in API specifications.")

    try:
        subnet = get_subnet_info(create_subnet(connection, **params))
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)

    # Sometimes AWS takes its time to create a subnet and so using
    # new subnets's id to do things like create tags results in
    # exception.
    if wait and subnet.get("state") != "available":
        wait_vpc_subnet_state(connection, module, "subnet_exists", start_time=start_time, SubnetIds=[subnet["id"]])
        wait_vpc_subnet_state(connection, module, "subnet_available", start_time=start_time, SubnetIds=[subnet["id"]])
        subnet["state"] = "available"

    return subnet


def disassociate_ipv6_cidr(conn, module: AnsibleAWSModule, subnet: Dict[str, Any], start_time: float):
    if subnet.get("assign_ipv6_address_on_creation"):
        try:
            modify_subnet_attribute(conn, subnet["id"], AssignIpv6AddressOnCreation={"Value": False})
        except AnsibleEC2Error as e:
            module.fail_json_aws_error(e)

    try:
        disassociate_subnet_cidr_block(conn, subnet["ipv6_association_id"])
    except AnsibleEC2Error as e:
        module.fail_json_aws(
            e,
            msg=f"Couldn't disassociate ipv6 cidr block id {subnet['ipv6_association_id']} from subnet {subnet['id']}",
        )

    # Wait for cidr block to be disassociated
    if module.params["wait"]:
        filters = ansible_dict_to_boto3_filter_list(
            {"ipv6-cidr-block-association.state": ["disassociated"], "vpc-id": subnet["vpc_id"]}
        )
        wait_vpc_subnet_state(
            conn, module, "subnet_exists", start_time=start_time, SubnetIds=[subnet["id"]], Filters=filters
        )


def ensure_ipv6_cidr_block(conn, module: AnsibleAWSModule, subnet: Dict[str, Any], start_time: float) -> bool:
    ipv6_cidr = module.params["ipv6_cidr"]
    wait = module.params["wait"]
    changed = False

    if subnet["ipv6_association_id"] and not ipv6_cidr:
        if not module.check_mode:
            disassociate_ipv6_cidr(conn, module, subnet, start_time)
        changed = True

    if ipv6_cidr:
        filters = ansible_dict_to_boto3_filter_list(
            {"ipv6-cidr-block-association.ipv6-cidr-block": ipv6_cidr, "vpc-id": subnet["vpc_id"]}
        )

        try:
            check_subnets = [get_subnet_info(x) for x in describe_subnets(client=conn, Filters=filters)]
        except AnsibleEC2Error as e:
            module.fail_json_aws_error(e)

        if check_subnets and check_subnets[0]["ipv6_cidr_block"]:
            module.fail_json(msg=f"The IPv6 CIDR '{ipv6_cidr}' conflicts with another subnet")

        if subnet["ipv6_association_id"]:
            if not module.check_mode:
                disassociate_ipv6_cidr(conn, module, subnet, start_time)
            changed = True

        try:
            if not module.check_mode:
                associate_resp = associate_subnet_cidr_block(conn, subnet_id=subnet["id"], Ipv6CidrBlock=ipv6_cidr)
            changed = True
        except AnsibleEC2Error as e:
            module.fail_json_aws(e, msg=f"Couldn't associate ipv6 cidr {ipv6_cidr} to {subnet['id']}")
        else:
            if not module.check_mode and wait:
                filters = ansible_dict_to_boto3_filter_list(
                    {"ipv6-cidr-block-association.state": ["associated"], "vpc-id": subnet["vpc_id"]}
                )
                wait_vpc_subnet_state(
                    conn, module, "subnet_exists", start_time=start_time, SubnetIds=[subnet["id"]], Filters=filters
                )
        if not module.check_mode:
            if associate_resp.get("AssociationId"):
                subnet["ipv6_association_id"] = associate_resp["AssociationId"]
                subnet["ipv6_cidr_block"] = associate_resp["Ipv6CidrBlock"]
                if subnet["ipv6_cidr_block_association_set"]:
                    subnet["ipv6_cidr_block_association_set"][0] = camel_dict_to_snake_dict(associate_resp)
                else:
                    subnet["ipv6_cidr_block_association_set"].append(camel_dict_to_snake_dict(associate_resp))

    return changed


def _matching_subnet_filters(vpc_id: str, cidr: List[str]) -> Dict[str, Union[str, List[str]]]:
    return ansible_dict_to_boto3_filter_list({"vpc-id": vpc_id, "cidr-block": cidr})


def get_matching_subnet(conn, module: AnsibleAWSModule, vpc_id: str, cidr: List[str]) -> Optional[Dict[str, Any]]:
    filters = _matching_subnet_filters(vpc_id, cidr)
    try:
        for subnet in describe_subnets(client=conn, Filters=filters):
            # The list contains only one subnet
            return get_subnet_info(subnet)
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)
    return None


def update_subnet_attributes(
    conn, module: AnsibleAWSModule, subnet: Dict[str, Any], start_time: float
) -> Tuple[bool, Dict[str, Any]]:
    changed = False
    # Ensure IPv6 CIDR Block
    if module.params["ipv6_cidr"] != subnet.get("ipv6_cidr_block"):
        changed |= ensure_ipv6_cidr_block(conn, module, subnet, start_time)

    # Modify subnet attribute 'MapPublicIpOnLaunch'
    map_public = module.params["map_public"]
    if map_public != subnet.get("map_public_ip_on_launch"):
        changed = True
        if not module.check_mode:
            params = {"MapPublicIpOnLaunch": {"Value": map_public}}
            modify_subnet_attribute(conn, subnet["id"], **params)

    # Modify subnet attribute 'AssignIpv6AddressOnCreation'
    assign_instances_ipv6 = module.params["assign_instances_ipv6"]
    if assign_instances_ipv6 != subnet.get("assign_ipv6_address_on_creation"):
        changed = True
        if not module.check_mode:
            params = {"AssignIpv6AddressOnCreation": {"Value": assign_instances_ipv6}}
            modify_subnet_attribute(conn, subnet["id"], **params)

    # Ensure subnet tags
    tags_updated = ensure_ec2_tags(
        conn,
        module,
        subnet["id"],
        resource_type="subnet",
        purge_tags=module.params["purge_tags"],
        tags=module.params["tags"],
        retry_codes=["InvalidSubnetID.NotFound"],
    )
    if tags_updated:
        changed = True
        if module.params["wait"] and not module.check_mode:
            # Wait for tags to be updated
            filters = ansible_dict_to_boto3_filter_list(ansible_dict_to_tag_filter_dict(module.params["tags"]))
            wait_vpc_subnet_state(
                conn, module, "subnet_exists", start_time=start_time, SubnetIds=[subnet["id"]], Filters=filters
            )

    subnet = get_matching_subnet(conn, module, module.params["vpc_id"], module.params["cidr"])
    if not module.check_mode and module.params["wait"]:
        subnet_filter = _matching_subnet_filters(module.params["vpc_id"], module.params["cidr"])
        wait_vpc_subnet_state(
            conn, module, "subnet_exists", start_time=start_time, SubnetIds=[subnet["id"]], Filters=subnet_filter
        )
        subnet = get_matching_subnet(conn, module, module.params["vpc_id"], module.params["cidr"])
        if not subnet:
            module.fail_json("Failed to describe newly created subnet")
        # GET calls are not monotonic for map_public_ip_on_launch and assign_ipv6_address_on_creation
        # so we only wait for those if necessary just before returning the subnet
        subnet = ensure_final_subnet(conn, module, subnet, start_time)
    return changed, subnet


def ensure_subnet_present(conn, module: AnsibleAWSModule) -> Dict[str, Any]:
    subnet = get_matching_subnet(conn, module, module.params["vpc_id"], module.params["cidr"])
    changed = False

    # Initialize start so max time does not exceed the specified wait_timeout for multiple operations
    start_time = time.time()

    if subnet is None:
        if not module.check_mode:
            subnet = create_vpc_subnet(conn, module, start_time)
        changed = True
        # Subnet will be None when check_mode is true
        if subnet is None:
            return {"changed": changed, "subnet": {}}

    attribute_change, subnet = update_subnet_attributes(conn, module, subnet, start_time)
    return {"changed": changed or attribute_change, "subnet": subnet}


def ensure_final_subnet(conn, module: AnsibleAWSModule, subnet: Dict[str, Any], start_time: float) -> Dict[str, Any]:
    for _rewait in range(0, 30):
        map_public_correct = True
        assign_ipv6_correct = True

        if module.params["map_public"] != subnet["map_public_ip_on_launch"]:
            map_public_correct = False
            waiter_name = "subnet_has_map_public" if module.params["map_public"] else "subnet_no_map_public"
            wait_vpc_subnet_state(conn, module, waiter_name, start_time=start_time, SubnetIds=[subnet["id"]])

        if module.params["assign_instances_ipv6"] != subnet.get("assign_ipv6_address_on_creation"):
            assign_ipv6_correct = False
            waiter_name = (
                "subnet_has_assign_ipv6" if module.params["assign_instances_ipv6"] else "subnet_no_assign_ipv6"
            )
            wait_vpc_subnet_state(conn, module, waiter_name, start_time=start_time, SubnetIds=[subnet["id"]])

        if map_public_correct and assign_ipv6_correct:
            break

        time.sleep(5)
        subnet = get_matching_subnet(conn, module, module.params["vpc_id"], module.params["cidr"]) or {}

    return subnet


def ensure_subnet_absent(conn, module: AnsibleAWSModule) -> Dict[str, Any]:
    subnet = get_matching_subnet(conn, module, module.params["vpc_id"], module.params["cidr"])
    if subnet is None:
        return {"changed": False}

    try:
        changed = False
        if module.check_mode:
            changed = True
        else:
            changed = delete_subnet(conn, subnet["id"])
            if module.params["wait"]:
                wait_vpc_subnet_state(conn, module, "subnet_deleted", start_time=time.time(), SubnetIds=[subnet["id"]])
    except AnsibleEC2Error as e:
        module.fail_json_aws_error(e)
    return {"changed": changed}


def main() -> None:
    argument_spec = dict(
        az=dict(default=None, required=False),
        cidr=dict(required=True),
        ipv6_cidr=dict(default="", required=False),
        outpost_arn=dict(default="", type="str", required=False),
        state=dict(default="present", choices=["present", "absent"]),
        tags=dict(required=False, type="dict", aliases=["resource_tags"]),
        vpc_id=dict(required=True),
        map_public=dict(default=False, required=False, type="bool"),
        assign_instances_ipv6=dict(default=False, required=False, type="bool"),
        wait=dict(type="bool", default=True),
        wait_timeout=dict(type="int", default=300, required=False),
        purge_tags=dict(default=True, type="bool"),
    )

    required_if = [("assign_instances_ipv6", True, ["ipv6_cidr"])]

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True, required_if=required_if)

    if module.params.get("outpost_arn") and not module.params.get("az"):
        module.fail_json(msg="To specify OutpostArn, you must specify the Availability Zone of the Outpost subnet.")

    if module.params.get("assign_instances_ipv6") and not module.params.get("ipv6_cidr"):
        module.fail_json(msg="assign_instances_ipv6 is True but ipv6_cidr is None or an empty string")

    connection = module.client("ec2")

    state = module.params.get("state")

    try:
        if state == "present":
            result = ensure_subnet_present(connection, module)
        elif state == "absent":
            result = ensure_subnet_absent(connection, module)
    except AnsibleEC2Error as e:
        module.fail_json_aws(e)

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2024, Björn Bösel <bjoernboesel@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_component

short_description: Allows administration of Keycloak components using Keycloak API

version_added: 10.0.0

description:
  - This module allows the administration of Keycloak components using the Keycloak REST API. It requires access to the REST
    API using OpenID Connect; the user connecting and the realm being used must have the requisite access rights. In a default
    Keycloak installation, C(admin-cli) and an C(admin) user would work, as would a separate realm definition with the scope
    tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/latest/rest-api/index.html). Aliases are provided so camelCased versions can be
    used as well.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the Keycloak component.
      - On V(present), the component is created (or updated if it exists already).
      - On V(absent), the component is removed if it exists.
    choices: ['present', 'absent']
    default: 'present'
    type: str
  name:
    description:
      - Name of the component to create.
    type: str
    required: true
  parent_id:
    description:
      - The parent_id of the component. In practice the ID (name) of the realm.
    type: str
    required: true
  provider_id:
    description:
      - The name of the "provider ID" for the key.
    type: str
    required: true
  provider_type:
    description:
      - The name of the "provider type" for the key. That is, V(org.keycloak.storage.UserStorageProvider), V(org.keycloak.userprofile.UserProfileProvider),
        ...
      - See U(https://www.keycloak.org/docs/latest/server_development/index.html#_providers).
    type: str
    required: true
  config:
    description:
      - Configuration properties for the provider.
      - Contents vary depending on the provider type.
    type: dict

extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Björn Bösel (@fivetide)
"""

EXAMPLES = r"""
- name: Manage Keycloak User Storage Provider
  community.general.keycloak_component:
    auth_keycloak_url: http://localhost:8080/auth
    auth_username: keycloak
    auth_password: keycloak
    auth_realm: master
    name: my storage provider
    state: present
    parent_id: some_realm
    provider_id: my storage
    provider_type: "org.keycloak.storage.UserStorageProvider"
    config:
      myCustomKey: "my_custom_key"
      cachePolicy: "NO_CACHE"
      enabled: true
"""

RETURN = r"""
end_state:
  description: Representation of the keycloak_component after module execution.
  returned: on success
  type: dict
  contains:
    id:
      description: ID of the component.
      type: str
      returned: when O(state=present)
      sample: 5b7ec13f-99da-46ad-8326-ab4c73cf4ce4
    name:
      description: Name of the component.
      type: str
      returned: when O(state=present)
      sample: mykey
    parentId:
      description: ID of the realm this key belongs to.
      type: str
      returned: when O(state=present)
      sample: myrealm
    providerId:
      description: The ID of the key provider.
      type: str
      returned: when O(state=present)
      sample: rsa
    providerType:
      description: The type of provider.
      type: str
      returned: when O(state=present)
    config:
      description: Component configuration.
      type: dict
"""

from copy import deepcopy
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    camel,
    get_token,
    keycloak_argument_spec,
)


def main():
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        state=dict(type="str", default="present", choices=["present", "absent"]),
        name=dict(type="str", required=True),
        parent_id=dict(type="str", required=True),
        provider_id=dict(type="str", required=True),
        provider_type=dict(type="str", required=True),
        config=dict(
            type="dict",
        ),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"]]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", end_state={}, diff=dict(before={}, after={}))

    # This will include the current state of the component if it is already
    # present. This is only used for diff-mode.
    before_component = {}
    before_component["config"] = {}

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    params_to_ignore = list(keycloak_argument_spec().keys()) + ["state", "parent_id"]

    # Filter and map the parameters names that apply to the role
    component_params = [x for x in module.params if x not in params_to_ignore and module.params.get(x) is not None]

    provider_type = module.params.get("provider_type")

    # Build a proposed changeset from parameters given to this module
    changeset = {}
    changeset["config"] = {}

    # Generate a JSON payload for Keycloak Admin API from the module
    # parameters.  Parameters that do not belong to the JSON payload (e.g.
    # "state" or "auth_keycloal_url") have been filtered away earlier (see
    # above).
    #
    # This loop converts Ansible module parameters (snake-case) into
    # Keycloak-compatible format (camel-case). For example private_key
    # becomes privateKey.
    #
    # It also converts bool, str and int parameters into lists with a single
    # entry of 'str' type. Bool values are also lowercased. This is required
    # by Keycloak.
    #
    for component_param in component_params:
        if component_param == "config":
            for config_param in module.params.get("config"):
                changeset["config"][camel(config_param)] = []
                raw_value = module.params.get("config")[config_param]
                if isinstance(raw_value, bool):
                    value = str(raw_value).lower()
                else:
                    value = str(raw_value)

                changeset["config"][camel(config_param)].append(value)
        else:
            # No need for camelcase in here as these are one word parameters
            new_param_value = module.params.get(component_param)
            changeset[camel(component_param)] = new_param_value

    # Make a deep copy of the changeset. This is use when determining
    # changes to the current state.
    changeset_copy = deepcopy(changeset)

    # Make it easier to refer to current module parameters
    name = module.params.get("name")
    state = module.params.get("state")
    provider_type = module.params.get("provider_type")
    parent_id = module.params.get("parent_id")

    # Get a list of all Keycloak components that are of keyprovider type.
    current_components = kc.get_components(urlencode(dict(type=provider_type)), parent_id)

    # If this component is present get its key ID. Confusingly the key ID is
    # also known as the Provider ID.
    component_id = None

    # Track individual parameter changes
    changes = ""

    # This tells Ansible whether the key was changed (added, removed, modified)
    result["changed"] = False

    # Loop through the list of components. If we encounter a component whose
    # name matches the value of the name parameter then assume the key is
    # already present.
    for component in current_components:
        if component["name"] == name:
            component_id = component["id"]
            changeset["id"] = component_id
            changeset_copy["id"] = component_id

            # Compare top-level parameters
            for param in changeset:
                before_component[param] = component[param]

                if changeset_copy[param] != component[param] and param != "config":
                    changes += f"{param}: {component[param]} -> {changeset_copy[param]}, "
                    result["changed"] = True
            # Compare parameters under the "config" key
            for p, v in changeset_copy["config"].items():
                try:
                    before_component["config"][p] = component["config"][p] or []
                except KeyError:
                    before_component["config"][p] = []
                if v != component["config"][p]:
                    changes += f"config.{p}: {component['config'][p]} -> {v}, "
                    result["changed"] = True

    # Check all the possible states of the resource and do what is needed to
    # converge current state with desired state (create, update or delete
    # the key).
    if component_id and state == "present":
        if result["changed"]:
            if module._diff:
                result["diff"] = dict(before=before_component, after=changeset_copy)

            if module.check_mode:
                result["msg"] = f"Component {name} would be changed: {changes.strip(', ')}"
            else:
                kc.update_component(changeset, parent_id)
                result["msg"] = f"Component {name} changed: {changes.strip(', ')}"
        else:
            result["msg"] = f"Component {name} was in sync"

        result["end_state"] = changeset_copy
    elif component_id and state == "absent":
        if module._diff:
            result["diff"] = dict(before=before_component, after={})

        if module.check_mode:
            result["changed"] = True
            result["msg"] = f"Component {name} would be deleted"
        else:
            kc.delete_component(component_id, parent_id)
            result["changed"] = True
            result["msg"] = f"Component {name} deleted"

        result["end_state"] = {}
    elif not component_id and state == "present":
        if module._diff:
            result["diff"] = dict(before={}, after=changeset_copy)

        if module.check_mode:
            result["changed"] = True
            result["msg"] = f"Component {name} would be created"
        else:
            kc.create_component(changeset, parent_id)
            result["changed"] = True
            result["msg"] = f"Component {name} created"

        result["end_state"] = changeset_copy
    elif not component_id and state == "absent":
        result["changed"] = False
        result["msg"] = f"Component {name} not present"
        result["end_state"] = {}

    module.exit_json(**result)


if __name__ == "__main__":
    main()
